"""Policies for the PursuitEvasion environment."""
import os.path as osp

from posggym.agents.grid_world.pursuit_evasion.shortest_path import PEShortestPathPolicy
from posggym.agents.registration import PolicySpec
from posggym.agents.torch_policy import PPOPolicy
from posggym.agents.utils import processors
from posggym.config import AGENT_MODEL_DIR

agent_model_dir = osp.join(AGENT_MODEL_DIR, "grid_world", "pursuit_evasion")
policy_specs = {}

spec = PolicySpec(
    policy_name="ShortestPath",
    entry_point=PEShortestPathPolicy,
    version=0,
    env_id="PursuitEvasion-v1",
    env_args=None,
    valid_agent_ids=None,
    nondeterministic=False,
    description=(
        "Takes the shortest path to the evader's goal (evader) or the evader's start "
        "location then the other possible evader start and goal locations (pursuer)."
    ),
)
policy_specs[spec.id] = spec


def _get_policy_description(policy_file_name):
    if policy_file_name.startswith("RL"):
        return "Deep RL policy trained using PPO and self-play."
    k = policy_file_name.split("_")[0].replace("KLR", "")
    if k == "BR":
        return (
            "Best-response to K-Level Reasoning policies. This is a deep RL policy "
            "training"
            " using PPO and the Synchronous KLR algorithm."
        )
    return (
        f"Level {k} K-Level Reasoning deep RL policy training using PPO and the "
        "Synchronous KLR algorithm."
    )


# PursuitEvasion 16x16
# Evader (agent=0)
for policy_file_name in [
    "KLR0_i0.pkl",
    "KLR1_i0.pkl",
    "KLR2_i0.pkl",
    "KLR3_i0.pkl",
    "KLR4_i0.pkl",
    "KLRBR_i0.pkl",
    "RL1_i0.pkl",
    "RL2_i0.pkl",
    "RL3_i0.pkl",
    "RL4_i0.pkl",
    "RL5_i0.pkl",
    "RL6_i0.pkl",
    "RL7_i0.pkl",
    "RL8_i0.pkl",
    "RL9_i0.pkl",
    "RL10_i0.pkl",
    "RL11_i0.pkl",
]:
    spec = PPOPolicy.get_spec_from_path(
        env_id="PursuitEvasion-v1",
        env_args={
            "grid": "16x16",
            "max_obs_distance": 12,
            "use_progress_reward": True,
        },
        env_args_id="grid=16x16",
        policy_file_path=osp.join(agent_model_dir, "grid=16x16", policy_file_name),
        version=0,
        valid_agent_ids=["0"],
        # policy is deterministic given random seed
        nondeterministic=False,
        # actions sampled, rather always taking most probable action
        deterministic=False,
        obs_processor_cls=processors.FlattenProcessor,
        obs_processor_config=None,
        action_processor_cls=None,
        action_processor_config=None,
        description=_get_policy_description(policy_file_name),
    )
    policy_specs[spec.id] = spec


# Pursuer (agent=1)
for policy_file_name in [
    "KLR0_i1.pkl",
    "KLR1_i1.pkl",
    "KLR2_i1.pkl",
    "KLR3_i1.pkl",
    "KLR4_i1.pkl",
    "KLRBR_i1.pkl",
    "RL1_i1.pkl",
    "RL2_i1.pkl",
    "RL3_i1.pkl",
    "RL4_i1.pkl",
    "RL5_i1.pkl",
    "RL6_i1.pkl",
    "RL7_i1.pkl",
    "RL8_i1.pkl",
    "RL9_i1.pkl",
    "RL10_i1.pkl",
    "RL11_i1.pkl",
]:
    spec = PPOPolicy.get_spec_from_path(
        env_id="PursuitEvasion-v1",
        env_args={
            "grid": "16x16",
            "max_obs_distance": 12,
            "use_progress_reward": True,
        },
        env_args_id="grid=16x16",
        policy_file_path=osp.join(agent_model_dir, "grid=16x16", policy_file_name),
        version=0,
        valid_agent_ids=["1"],
        # policy is deterministic given random seed
        nondeterministic=False,
        # actions sampled, rather than always taking most probable action
        deterministic=False,
        obs_processor_cls=processors.FlattenProcessor,
        obs_processor_config=None,
        action_processor_cls=None,
        action_processor_config=None,
        description=_get_policy_description(policy_file_name),
    )
    policy_specs[spec.id] = spec

# PursuitEvasion 8x8
# Evader (agent=0)
for policy_file_name in [
    "KLR0_i0.pkl",
    "KLR1_i0.pkl",
    "KLR2_i0.pkl",
    "KLR3_i0.pkl",
    "KLR4_i0.pkl",
    "KLRBR_i0.pkl",
    "RL1_i0.pkl",
    "RL2_i0.pkl",
    "RL3_i0.pkl",
    "RL4_i0.pkl",
    "RL5_i0.pkl",
    "RL6_i0.pkl",
    "RL7_i0.pkl",
]:
    spec = PPOPolicy.get_spec_from_path(
        env_id="PursuitEvasion-v1",
        env_args={
            "grid": "8x8",
            "max_obs_distance": 12,
            "use_progress_reward": True,
        },
        env_args_id="grid=8x8",
        policy_file_path=osp.join(agent_model_dir, "grid=8x8", policy_file_name),
        version=0,
        valid_agent_ids=["0"],
        # policy is deterministic given random seed
        nondeterministic=False,
        # actions sampled, rather always taking most probable action
        deterministic=False,
        obs_processor_cls=processors.FlattenProcessor,
        obs_processor_config=None,
        action_processor_cls=None,
        action_processor_config=None,
        description=_get_policy_description(policy_file_name),
    )
    policy_specs[spec.id] = spec


# Pursuer (agent=1)
for policy_file_name in [
    "KLR0_i1.pkl",
    "KLR1_i1.pkl",
    "KLR2_i1.pkl",
    "KLR3_i1.pkl",
    "KLR4_i1.pkl",
    "KLRBR_i1.pkl",
    "RL1_i1.pkl",
    "RL2_i1.pkl",
    "RL3_i1.pkl",
    "RL4_i1.pkl",
    "RL5_i1.pkl",
    "RL6_i1.pkl",
    "RL7_i1.pkl",
]:
    spec = PPOPolicy.get_spec_from_path(
        env_id="PursuitEvasion-v1",
        env_args={
            "grid": "8x8",
            "max_obs_distance": 12,
            "use_progress_reward": True,
        },
        env_args_id="grid=8x8",
        policy_file_path=osp.join(agent_model_dir, "grid=8x8", policy_file_name),
        version=0,
        valid_agent_ids=["1"],
        # policy is deterministic given random seed
        nondeterministic=False,
        # actions sampled, rather always taking most probable action
        deterministic=False,
        obs_processor_cls=processors.FlattenProcessor,
        obs_processor_config=None,
        action_processor_cls=None,
        action_processor_config=None,
        description=_get_policy_description(policy_file_name),
    )
    policy_specs[spec.id] = spec



"""Environment wrapper class that stacks agent observations into a single array."""
from typing import Any, Dict

import numpy as np
from gymnasium import spaces

import posggym


class StackEnv(posggym.Wrapper):
    """Converts environment to handle single array stacked agent inputs and outputs.

    This wrapper converts the :meth:`step` and :meth:`reset` functions handle actions,
    observations, rewards, and termination signals of all agents as a single stacked
    array, as opposed to a dictionary with one array per agent.

    For example, if the observation space of the unwrapped environment is such that::

        >>> env.observation_spaces
        {
            "agent-1": Box(0.0, 1.0, (4,), float32),
            "agent-2": Box(0.0, 1.0, (4,), float32),
        }
        >>> acts = {"agent-1": 1, "agent-2": 2}
        >>> obs, rews, terms, truncs, all_done, infos = env.step(acts)
        >>> obs
        {
            "agent-1": np.array([0.1, 0.2, 0.3, 0.4]),
            "agent-2": np.array([0.5, 0.6, 0.7, 0.8]),
        }
        >>> obs["agent-1"].shape
        (4,)

    Wrapping this environment with the :class:`StackEnv` wrapper will change the output
    and input to the form::

        >>> stacked_acts = np.array([1, 2])
        >>> obs, rews, terms, truncs, all_done, infos = wrapped_env.step(stacked_acts)
        >>> obs
        np.array([[0.1, 0.2, 0.3, 0.4], [0.5, 0.6, 0.7, 0.8]])
        >>> obs.shape
        (2, 4)
        >>> rews.shape
        (2,)
        >>> terms.shape
        (2,)
        >>> truncs.shape
        (2,)

    Note, if the environment is a vector environment, the output will be stacked
    in order of the each sub-environment, as opposed to in order of each agent. For
    example, if we have a vector environment with two agents and two sub-environments,
    the observation space of the vector environment will be something like::

        >>> vec_env.num_envs
        2
        >>> env.single_observation_spaces
        {
            "agent-1": Box(0.0, 1.0, (4,), float32),
            "agent-2": Box(0.0, 1.0, (4,), float32),
        }
        >>> env.observation_spaces
        {
            "agent-1": Box(0.0, 1.0, (2, 4), float32),
            "agent-2": Box(0.0, 1.0, (2, 4), float32),
        }

    The output from :meth:`step` and :meth:`reset` of the StackEnv wrapper around the
    vector env will be in the form of::

        >>> obs, rewards, terms, truncs, all_done, infos = wrapped_env.step(actions)
        >>> obs.shape
        (4, 4)

    With the first two rows corresponding to the outputs for `agent-1` and `agent-2` in
    the first sub-environment and the last two rows corresponding the respective agents
    outputs in the second sub-environment.

    Similarly, the actions passed to the step method of the StackEnv wrapper around the
    vector env are expected to be stacked in order of each environment. For example, for
    the vector environment above, assuming discrete actions, the action will look like::

        >>> actions
        np.array([0, 1, 0, 1])

    With the first two elements corresponding to the actions for `agent-1` and `agent-2`
    in the first sub-environment and the last two elements corresponding to the actions
    for the respective agents in the second sub-environment.

    """

    def __init__(self, env: posggym.Env):
        super().__init__(env)

        self.num_envs = getattr(env, "num_envs", 1)
        self.is_vector_env = getattr(env, "is_vector_env", False)

        agent_0_obs_space = env.observation_spaces[env.possible_agents[0]]
        for obs_space in env.observation_spaces.values():
            assert isinstance(obs_space, spaces.Box), (
                f"StackEnv wrapper only supports environments with Box observation "
                f"spaces. Got {obs_space}. (Hint: try using the FlattenObservations "
                "wrapper.)"
            )
            assert obs_space == agent_0_obs_space, (
                f"StackEnv wrapper only supports environments where every agent "
                f"has the the same observation space. "
                f"Got {obs_space} and {agent_0_obs_space}."
            )

        agent_0_action_space = env.action_spaces[env.possible_agents[0]]
        for action_space in env.action_spaces.values():
            assert action_space == agent_0_action_space, (
                f"StackEnv wrapper only supports environments where every agent "
                f"has the the same action space. "
                f"Got {action_space} and {agent_0_action_space}."
            )

    def reset(self, **kwargs):
        obs, infos = super().reset(**kwargs)
        return self._stack_output(obs), infos

    def step(self, actions):
        # input shape (num_envs * num_agents, *single_action_space.shape)
        # convert to dict of actions, shape (num_envs, *single_action_space.shape)
        action_map = {
            i: actions[idx :: len(self.possible_agents)]
            for idx, i in enumerate(self.possible_agents)
        }
        obs, rewards, terminated, truncated, dones, infos = self.env.step(action_map)
        return (
            self._stack_output(obs),
            self._stack_output(rewards),
            self._stack_output(terminated),
            self._stack_output(truncated),
            dones if isinstance(dones, np.ndarray) else np.array([dones]),
            infos,
        )

    def _stack_output(self, output: Dict[str, Any]) -> np.ndarray:
        """Stacks the output of the environment into a single array."""
        x0 = list(output.values())[0]
        x0 = x0 if isinstance(x0, np.ndarray) else np.array([x0])
        num_agents = len(self.possible_agents)
        if self.is_vector_env:
            # agent outputs will have shape (num_envs, *single_env_output_shape)
            out_array = np.zeros((self.num_envs * num_agents, *x0.shape[1:]), x0.dtype)
            for i in output:
                idx = self.possible_agents.index(i)
                out_array[idx::num_agents] = output[i]
            return out_array

        # otherwise, agent output will have shape (single_env_output_shape, )
        x = [output.get(i, np.zeros(x0.shape, x0.dtype)) for i in self.possible_agents]
        return np.stack(x, axis=0)



"""Functionality for measuring return based diversity of posggym.agents policies."""
from __future__ import annotations

import os
from itertools import product
from typing import Dict, List

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

import posggym.agents as pga
from posggym.agents.evaluation import pairwise


def measure_return_diversity(
    pw_returns: np.ndarray,
    policies: List[str],
    co_teams: List[str],
    verbose: bool = False,
) -> Dict[str, np.ndarray]:
    """Measure return diversity for a set of pairwise returns.

    Diversity is measured in terms of the Euclidean distance between the pairwise
    return distributions of each policy.

    Returns
    -------
    pw_ed : np.ndarray
        The pairwise ED between return distributions of each policy.
    policy_ed : np.ndarray
        The total ED for each policy (the larger the more diverse).
    clusters : np.ndarray
        The cluster each policy belongs to. Cluster ordering arbitrary, but all policies
        in the same cluster are similar w.r.t ED between their return distributions.

    """
    assert (len(policies), len(co_teams)) == pw_returns.shape

    # get max and min returns, excluding random policy if it exists
    random_co_team_idxs = []
    random_idx = None
    if len(policies) > 2 and "Random" in policies:
        random_idx = policies.index("Random")
        pw_returns_excl_random = np.delete(pw_returns, random_idx, axis=0)
        random_co_team_idxs = [
            idx for idx, co_team in enumerate(co_teams) if "Random" in co_team
        ]
        pw_returns_excl_random = np.delete(
            pw_returns_excl_random, random_co_team_idxs, axis=1
        )
        max_return = np.max(pw_returns_excl_random)
        min_return = np.min(pw_returns_excl_random)
    else:
        max_return, min_return = np.max(pw_returns), np.min(pw_returns)

    if verbose:
        with np.printoptions(precision=2, suppress=True):
            # print(f"{pw_returns=}")
            print(f"{max_return=:.2f}, {min_return=:.2f}")

    # normalize pairwise distributions into [0.0, 1.0]
    pw_returns_norm = (pw_returns - min_return) / (max_return - min_return)
    if verbose:
        with np.printoptions(precision=2, suppress=True):
            for idx, policy_name in enumerate(policies):
                print(f"{policy_name}:\n{pw_returns_norm[idx]}")

    # compute ED between normalized return distributions of each policy
    pw_ed = np.zeros((len(policies), len(policies)))
    for pi_idx, pj_idx in product(range(len(policies)), repeat=2):
        pw_ed[pi_idx, pj_idx] = np.sqrt(
            np.sum((pw_returns[pi_idx] - pw_returns[pj_idx]) ** 2)
        )

    if verbose:
        print("Pairwise ED:")
        with np.printoptions(precision=2, suppress=True):
            for pi_idx, row in enumerate(pw_ed):
                print(policies[pi_idx], row)

    # Compute total CE for each policy (the larger the more diverse)
    policy_ed = np.zeros(len(policies), dtype=float)
    for pi_idx in range(len(policies)):
        policy_ed[pi_idx] = np.sum(pw_ed[pi_idx])

    if verbose:
        with np.printoptions(precision=2, suppress=True):
            print(f"{policy_ed=}")

    # Group similar policies by relative MSE
    if len(policies) > 2 and "Random" in policies:
        # exclude random policy from calculating bin sizes for grouping
        pw_ed_excl_random = np.delete(pw_ed, random_idx, axis=0)
        pw_ed_excl_random = np.delete(pw_ed_excl_random, random_idx, axis=1)
        min_ed, max_ed = np.min(pw_ed_excl_random), np.max(pw_ed_excl_random)
    else:
        min_ed, max_ed = np.min(pw_ed), np.max([pw_ed])
    threshold = min_ed + (max_ed - min_ed) * 0.1

    similar_policy_groups = [[policies[0]]]
    for policy_name in policies[1:]:
        pi_idx = policies.index(policy_name)
        min_div = float("inf")
        min_div_group_idx = 0
        for group_idx, group in enumerate(similar_policy_groups):
            pj_idx = policies.index(group[0])
            if pw_ed[pi_idx, pj_idx] < min_div:
                min_div = pw_ed[pi_idx, pj_idx]
                min_div_group_idx = group_idx

        if min_div < threshold:
            similar_policy_groups[min_div_group_idx].append(policy_name)
        else:
            similar_policy_groups.append([policy_name])

    if verbose:
        print("Similar policies:")
        print(f"{min_ed=} {max_ed=} {threshold=}")
        for group in similar_policy_groups:
            print(group)

    # Compute clusters
    clusters = np.zeros(len(policies), dtype=int)
    for cluster_idx, group in enumerate(similar_policy_groups):
        for policy_name in group:
            clusters[policies.index(policy_name)] = cluster_idx

    if verbose:
        with np.printoptions(precision=2, suppress=True):
            print(f"{clusters=}")

    return {
        "pairwise_return_ed": pw_ed,
        "policy_ed": policy_ed.reshape(-1, 1),
        "return_ed_clusters": clusters.reshape(-1, 1),
    }


def run_return_diversity_analysis(
    output_dir: str,
    env_id: str | None,
    env_args_id: str | None = None,
    verbose: bool = False,
):
    """Run return diversity analysis from pairwise comparison results.

    Results of analysis will be output to the results directory for each env.
    """
    all_envs = pga.get_all_envs()
    if env_id is None:
        env_ids = list(all_envs)
        env_args_id = None
    else:
        env_ids = [env_id]
        env_args_id = env_args_id

    for env_id in os.listdir(output_dir):
        if env_id not in env_ids:
            continue
        print(f"Running return diversity analysis for {env_id=}")
        env_results_dir = os.path.join(output_dir, env_id)
        print(f"  results saved to {env_results_dir=}")

        env_result_files = [
            fname for fname in os.listdir(env_results_dir) if fname.endswith(".csv")
        ]

        env_args_map = pga.get_all_envs()[env_id]
        for args_id in env_args_map:
            if (env_args_id is not None and args_id != env_args_id) or (
                env_args_id is None and str(args_id) + ".csv" not in env_result_files
            ):
                continue

            print(f"  {args_id=}")
            df = pairwise.load_pairwise_comparison_results(env_id, output_dir, args_id)

            if df["symmetric"].unique().tolist()[0]:
                # only keep results for one agent
                agent_ids = df["agent_id"].unique().tolist()
                df = df[df["agent_id"] == agent_ids[0]]

            pw_returns_per_agent = pairwise.get_pairwise_returns_matrix(df)

            pairwise.generate_pairwise_returns_plot(
                env_id=env_id,
                output_dir=env_results_dir,
                env_args_id=args_id,
                pw_returns_per_agent=pw_returns_per_agent,
                show=False,
                save=True,
            )

            results = {}
            for i in pw_returns_per_agent:
                pw_returns, policy_ids, co_teams_ids = pw_returns_per_agent[i]
                results_i = measure_return_diversity(
                    pw_returns[0],
                    policy_ids,
                    co_teams_ids,
                    verbose=verbose,
                )
                results[i] = (results_i, policy_ids, co_teams_ids)

            for k in list(results.values())[0][0]:
                fig, axs = plt.subplots(
                    nrows=1,
                    ncols=len(results),
                    figsize=(6 * len(results), 6),
                    squeeze=False,
                    layout="constrained",
                )
                for idx, i in enumerate(results):
                    div_results, policy_ids, co_teams_ids = results[i]

                    if len(results) > 2:
                        # only show first policy name for each co-team
                        co_team_labels = [
                            team_id.replace("(", "").replace(")", "").split(",")[0]
                            for team_id in co_teams_ids
                        ]
                    else:
                        co_team_labels = co_teams_ids

                    if div_results[k].shape[1] == 1:
                        xticklabels = [""]
                        annot_fontsize = 8
                    elif k in ["pairwise_return_ed"]:
                        xticklabels = policy_ids
                        annot_fontsize = min(8, 24 / np.sqrt(max(div_results[k].shape)))
                    else:
                        xticklabels = co_team_labels
                        annot_fontsize = min(8, 24 / np.sqrt(max(div_results[k].shape)))

                    sns.heatmap(
                        div_results[k],
                        ax=axs[0][idx],
                        annot=True,
                        fmt=".0f" if div_results[k].dtype == int else ".2f",
                        cmap="YlGnBu",
                        xticklabels=xticklabels,
                        yticklabels=policy_ids,
                        square=div_results[k].shape[1] != 1,
                        annot_kws={"fontsize": annot_fontsize},
                    )
                    axs[0][idx].set_title(f"agent `{i}`")
                fig.savefig(
                    os.path.join(env_results_dir, f"{args_id}_{k}.svg"),
                    bbox_inches="tight",
                )



"""The Policy class defining the core API for posggym.agents policies."""
from __future__ import annotations

import abc
import copy
from typing import TYPE_CHECKING, Any, Dict, Generic

from posggym.model import ActType, ObsType


if TYPE_CHECKING:
    from posggym.agents.registration import PolicySpec
    from posggym.agents.utils.action_distributions import ActionDistribution
    from posggym.model import POSGModel
    from posggym.utils.history import AgentHistory


# Convenient type definitions
PolicyID = str
PolicyState = Dict[str, Any]


class Policy(abc.ABC, Generic[ActType, ObsType]):
    """The main POSGGym Agents class implementing a policy for a single agent.

    This class defines the core POSGGym Agents policy API for a single agent acting in
    a partially observable, multi-agent (POSG) environment. The policy can be used to
    interact with an environment vis the :meth:`step` method, which returns the next
    action to take given the latest observation.

    The API is designed to be as general as possible while also following a similar
    design to the POSGGym Env API (which in turn follows the Gymnasium API).

    The main Agent Policy API methods that users of this class need to know are:

    - :meth:`step`
    - :meth:`reset`
    - :meth:`close`

    And the main attributes are:

    - :attr:`model`
    - :attr:`agent_id`
    - :attr:`policy_id`
    - :attr:`spec`
    - :attr:`observes_state`

    The policy API also supports finer grained control over the policy's internal state
    via the following methods:

    - :meth:`get_initial_state`
    - :meth:`get_next_state`
    - :meth:`sample_action`
    - :meth:`get_pi`
    - :meth:`get_value`
    - :meth:`set_state`
    - :meth:`get_state`
    - :meth:`get_state_from_history`

    The policy maintains an internal state, which is a dictionary and stored in the
    :attr:`_state` attribute. This state is updated each time the :meth:`step` method is
    called and reset to an initial state whenever the :meth:`reset` method is called.
    By default the policy state is an empty dictionary. However, subclasses should
    extend this state to include any additional information required by the policy to
    select the next action given it's action-observation history.

    The policy also maintains a reference to the last action it returned by the
    :meth:`step` method, which is stored in the :attr:`_last_action` attribute. This
    attribute is set to None when the policy is first created and whenever the
    :meth:`reset` method is called. It is used to update the policy's internal state
    each time the :meth:`step` method is called.


    """

    # PolicySpec used to initialize policy instance
    # This is set when policy is made using posggym.agents.make function
    spec: PolicySpec | None = None

    # Whether the policy expects the full state as it's observation or not
    observes_state: bool = False

    def __init__(self, model: POSGModel, agent_id: str, policy_id: PolicyID):
        self.model = model
        self.agent_id = agent_id
        self.policy_id = policy_id
        self._state = self.get_initial_state()
        self._last_action = None

    def step(self, obs: ObsType) -> ActType:
        """Get the next action from the policy.

        This function updates the policy's current internal state given the most recent
        observation, and returns the next action for the policy.

        Arguments
        ---------
        obs : ObsType
            the latest observation.

        Returns
        -------
        action : ActType
            the next action

        """
        self._state = self.get_next_state(self._last_action, obs, self._state)
        self._last_action = self.sample_action(self._state)
        return self._last_action

    def reset(self, *, seed: int | None = None):
        """Reset the policy to it's initial state.

        This resets the policy's internal state, and should be called at the start of
        each episode.

        Subclasses that use random number generators (RNG) should override this
        method and seed the RNG if a `seed` is provided. The expected behaviour is that
        this is that the seed provided once by the user, just after the policy is first
        created and before it interacts with an environment.

        Arguments
        ---------
        seed : int, optional
            seed for random number generator.

        """
        self._state = self.get_initial_state()
        self._last_action = None

    def close(self):
        """Close policy and perform any necessary cleanup.

        Should be overridden in subclasses as necessary.
        """
        pass

    def get_initial_state(self) -> PolicyState:
        """Get the policy's initial state.

        Subclasses that utilize custom internal states (e.g. RNN policies) should
        override this method, but should first call `super.().get_initial_state()` to
        get the base policy state that can then be extended.

        Returns
        -------
        initial_state : PolicyState
            the initial policy state

        """
        return {}

    @abc.abstractmethod
    def get_next_state(
        self,
        action: ActType | None,
        obs: ObsType,
        state: PolicyState,
    ) -> PolicyState:
        """Get the next policy state.

        Subclasses must implement this method.

        Arguments
        ---------
        action : ActType, optional
            the action performed. May be None if this is the first observation.
        obs : ObsType
            the observation received
        state : PolicyState
            the policy's state before action was performed and obs received

        Returns
        -------
        next_state : PolicyState
            the next policy state

        """

    @abc.abstractmethod
    def sample_action(self, state: PolicyState) -> ActType:
        """Sample an action given policy's current state.

        If the policy is deterministic then this will return the same action each time,
        given the same state. If the policy is stochastic then the action may change
        each time even if the state is the same.

        Subclasses must implement this method.

        Arguments
        ---------
        state : PolicyState
            the policy's current state

        Returns
        -------
        action : ActType
            the sampled action

        """

    @abc.abstractmethod
    def get_pi(self, state: PolicyState) -> ActionDistribution:
        """Get policy's distribution over actions for given policy state.

        Subclasses must implement this method, and should return an instance of a
        subclass of the
        :py:class:`posggym.agents.utils.action_distributions.ActionDistribution`
        class.

        Arguments
        ---------
        state : PolicyState
            the policy's current state

        Returns
        -------
        pi : ActionDistribution
            the policy's distribution over actions

        """

    @abc.abstractmethod
    def get_value(self, state: PolicyState) -> float:
        """Get a value estimate of a history.

        Subclasses must implement this method, but may set it to raise a
        NotImplementedError if the policy does not support value estimates.

        Arguments
        ---------
        state : PolicyState
            the policy's current state

        Returns
        -------
        value : float
            the value estimate

        """

    def set_state(self, state: PolicyState, last_action: ActType | None = None):
        """Set the policy's internal state.

        Subclasses that utilize custom internal states (e.g. RNN policies) may wish to
        override this method, to set any attributes used for the by the class to store
        policy state.

        Arguments
        ---------
        state : PolicyState
            the new policy state
        last_action : ActType, optional
            the last action taken by the policy. If not provided then the last action
            will be set to None.

        Raises
        ------
        AssertionError :
            if new policy state is not valid.

        """
        if self._state is not None and state is not None:
            assert all(k in state for k in self._state), f"Invalid policy state {state}"
        self._state = state
        self._last_action = last_action

    def get_state(self) -> PolicyState:
        """Get the policy's current state.

        Returns
        -------
        state : PolicyState
            policy's current internal state.

        """
        return copy.deepcopy(self._state)

    def get_state_from_history(self, history: AgentHistory) -> PolicyState:
        """Get the policy's state given history.

        This function essentially unrolls the policy using the actions and observations
        contained in the agent history.

        Note, this function will return None for the action in the final output state,
        as this would correspond to the action that was selected by the policy to action


        Arguments
        ---------
        history : AgentHistory
            the agent's action-observation history

        Returns
        -------
        state : PolicyState
            policy state given history

        """
        state = self.get_initial_state()
        for a, o in history:
            state = self.get_next_state(a, o, state)
        return state



"""A set of functions for passively checking environment implementations.

This module is heavily based on the passive_env_checker from gymnasium. It's copied and
adapted here to ensure compatibility across different gymnasium versions:
https://github.com/Farama-Foundation/Gymnasium/blob/v0.27.0/gymnasium/utils/passive_env_checker.py

"""
import inspect
import random
from functools import partial
from typing import Callable, Dict, Optional, Sequence

import numpy as np
from gymnasium import Space, spaces

import posggym
import posggym.model as M
from posggym import error, logger
from posggym.utils import seeding


def data_equivalence(data_1, data_2) -> bool:
    """Assert equality between data 1 and 2, i.e observations, actions, info.

    Arguments
    ---------
    data_1: data structure 1
    data_2: data structure 2

    Returns
    -------
    bool: If observation 1 and 2 are equivalent

    """
    if type(data_1) == type(data_2):
        if isinstance(data_1, dict):
            return data_1.keys() == data_2.keys() and all(
                data_equivalence(data_1[k], data_2[k]) for k in data_1
            )
        elif isinstance(data_1, (tuple, list)):
            return len(data_1) == len(data_2) and all(
                data_equivalence(o_1, o_2) for o_1, o_2 in zip(data_1, data_2)
            )
        elif isinstance(data_1, np.ndarray):
            return data_1.shape == data_2.shape and np.allclose(
                data_1, data_2, atol=0.00001
            )
        else:
            return data_1 == data_2
    else:
        return False


def check_rng_equality(rng_1: seeding.RNG, rng_2: seeding.RNG, prefix=None):
    """Check equality between two random number generators."""
    assert type(rng_1) == type(
        rng_2
    ), f"{prefix}Differing RNG types: {rng_1} and {rng_2}"
    if isinstance(rng_1, random.Random) and isinstance(rng_2, random.Random):
        assert (
            rng_1.getstate() == rng_2.getstate()
        ), f"{prefix}Internal states differ: {rng_1} and {rng_2}"
    elif isinstance(rng_1, np.random.Generator) and isinstance(
        rng_2, np.random.Generator
    ):
        assert (
            rng_1.bit_generator.state == rng_2.bit_generator.state
        ), f"{prefix}Internal states differ: {rng_1} and {rng_2}"
    else:
        raise AssertionError(f"{prefix}Unsupported RNG type: '{type(rng_1)}'.")


def check_space_limit(space, space_type: str):
    """Check the space limit for only the Box space."""
    if isinstance(space, spaces.Box):
        if np.any(np.equal(space.low, -np.inf)):
            logger.warn(
                f"A Box {space_type} space minimum value is -infinity. This is "
                "probably too low."
            )
        if np.any(np.equal(space.high, np.inf)):
            logger.warn(
                f"A Box {space_type} space maximum value is -infinity. This is "
                "probably too high."
            )
        # Check that the Box vector space is normalized
        if (
            space_type == "action"
            and len(space.shape) == 1
            and (
                np.any(
                    np.logical_and(
                        space.low != np.zeros_like(space.low),
                        np.abs(space.low) != np.abs(space.high),
                    )
                )
                or np.any(space.low < -1)
                or np.any(space.high > 1)
            )
        ):
            logger.warn(
                "For Box action spaces, we recommend using a symmetric and normalized "
                "space (range=[-1, 1] or [0, 1]). See "
                "https://stable-baselines3.readthedocs.io/en/master/guide/rl_tips.html "
                "for more information."
            )
    elif isinstance(space, spaces.Tuple):
        for subspace in space.spaces:
            check_space_limit(subspace, space_type)
    elif isinstance(space, spaces.Dict):
        for subspace in space.values():
            check_space_limit(subspace, space_type)


def check_agent_space_limits(agent_spaces: Dict[str, Space], space_type: str):
    """Check the space limit for any agent Box space."""
    for i, agent_space in agent_spaces.items():
        check_space_limit(agent_space, space_type)


def _check_box_state_space(state_space: spaces.Box):
    """Checks that a :class:`Box` state space is defined in a sensible way."""
    assert state_space.low.shape == state_space.shape, (
        "The Box state space shape and low shape have have different shapes, "
        f"low shape: {state_space.low.shape}, box shape: {state_space.shape}"
    )
    assert state_space.high.shape == state_space.shape, (
        f"The Box state space shape and high shape have different shapes, "
        f"high shape: {state_space.high.shape}, box shape: {state_space.shape}"
    )

    if np.any(state_space.low == state_space.high):
        logger.warn(
            "A Box state space maximum and minimum values are equal. "
            "Actual equal coordinates: "
            f"{list(zip(*np.where(state_space.low == state_space.high)))}"
        )
    elif np.any(state_space.high < state_space.low):
        logger.warn(
            "A Box state space low value is greater than a high value. "
            "Actual less than coordinates: "
            f"{list(zip(*np.where(state_space.high < state_space.low)))}"
        )


def _check_box_observation_space(observation_space: spaces.Box):
    """Checks that a :class:`Box` observation space is defined in a sensible way."""
    # Check if the box is an image
    if (len(observation_space.shape) == 3 and observation_space.shape[0] != 1) or (
        len(observation_space.shape) == 4 and observation_space.shape[0] == 1
    ):
        if observation_space.dtype != np.uint8:
            logger.warn(
                "It seems a Box observation space is an image but the `dtype` is not "
                f"`np.uint8`, actual type: {observation_space.dtype}. "
                "If the Box observation space is not an image, we recommend flattening "
                "the observation to have only a 1D vector."
            )
        if np.any(observation_space.low != 0) or np.any(observation_space.high != 255):
            logger.warn(
                "It seems a Box observation space is an image but the lower and upper "
                "bounds are not [0, 255]. "
                f"Actual lower bound: {np.min(observation_space.low)}, upper bound: "
                f"{np.max(observation_space.high)}. "
                "Generally, CNN policies assume observations are within that range, so "
                "you may encounter an issue if the observation values are not."
            )

    if len(observation_space.shape) not in [1, 3] and not (
        len(observation_space.shape) == 2 and observation_space.shape[0] == 1
    ):
        logger.warn(
            "A Box observation space has an unconventional shape (neither an image, "
            "nor a 1D vector). We recommend flattening the observation to have only a "
            "1D vector or use a custom policy to properly process the data. "
            f"Actual observation shape: {observation_space.shape}"
        )

    assert observation_space.low.shape == observation_space.shape, (
        f"The Box observation space shape and low shape have different shapes, low "
        f"shape: {observation_space.low.shape}, box shape: {observation_space.shape}"
    )
    assert observation_space.high.shape == observation_space.shape, (
        "The Box observation space shape and high shape have have different shapes, "
        f"high shape: {observation_space.high.shape}, box shape: "
        f"{observation_space.shape}"
    )

    if np.any(observation_space.low == observation_space.high):
        logger.warn(
            "A Box observation space maximum and minimum values are equal. "
            "Actual equal coordinates: "
            f"{list(zip(*np.where(observation_space.low == observation_space.high)))}"
        )
    elif np.any(observation_space.high < observation_space.low):
        logger.warn(
            "A Box observation space low value is greater than a high value. "
            "Actual less than coordinates: "
            f"{list(zip(*np.where(observation_space.high < observation_space.low)))}"
        )


def _check_box_action_space(action_space: spaces.Box):
    """Checks that a :class:`Box` action space is defined in a sensible way."""
    assert action_space.low.shape == action_space.shape, (
        f"The Box action space shape and low shape have have different shapes, low "
        f"shape: {action_space.low.shape}, box shape: {action_space.shape}"
    )
    assert action_space.high.shape == action_space.shape, (
        f"The Box action space shape and high shape have different shapes, high shape: "
        f"{action_space.high.shape}, box shape: {action_space.shape}"
    )

    if np.any(action_space.low == action_space.high):
        logger.warn(
            "A Box action space maximum and minimum values are equal. "
            "Actual equal coordinates: "
            f"{list(zip(*np.where(action_space.low == action_space.high)))}"
        )
    elif np.any(action_space.high < action_space.low):
        logger.warn(
            "A Box action space low value is greater than a high value. "
            "Actual less than coordinates: "
            f"{list(zip(*np.where(action_space.high < action_space.low)))}"
        )


def check_space(
    space: Space, space_type: str, check_box_space_fn: Callable[[spaces.Box], None]
):
    """A passive check of an environment's space."""
    if not isinstance(space, spaces.Space):
        raise AssertionError(
            f"{space_type} space does not inherit from `gymnasium.spaces.Space`, "
            f"actual type: {type(space)}"
        )

    elif isinstance(space, spaces.Box):
        check_box_space_fn(space)
    elif isinstance(space, spaces.Discrete):
        assert space.n > 0, (
            f"Discrete {space_type} space's number of elements must be positive, "
            f"actual number of elements: {space.n}"
        )
        assert space.shape == (), (
            f"Discrete {space_type} space's shape should be empty, actual shape: "
            f"{space.shape}"
        )
    elif isinstance(space, spaces.MultiDiscrete):
        assert space.shape == space.nvec.shape, (
            f"Multi-discrete {space_type} space's shape must be equal to the nvec "
            f"shape, space shape: {space.shape}, nvec shape: {space.nvec.shape}"
        )
        assert np.all(space.nvec > 0), (
            f"Multi-discrete {space_type} space's all nvec elements must be greater "
            f"than 0, actual nvec: {space.nvec}"
        )
    elif isinstance(space, spaces.MultiBinary):
        assert np.all(np.asarray(space.shape) > 0), (
            f"Multi-binary {space_type} space's all shape elements must be greater "
            f"than 0, actual shape: {space.shape}"
        )
    elif isinstance(space, spaces.Tuple):
        assert (
            len(space.spaces) > 0
        ), f"An empty Tuple {space_type} space is not allowed."
        for subspace in space.spaces:
            check_space(subspace, space_type, check_box_space_fn)
    elif isinstance(space, spaces.Dict):
        assert (
            len(space.spaces.keys()) > 0
        ), f"An empty Dict {space_type} space is not allowed."
        for subspace in space.values():
            check_space(subspace, space_type, check_box_space_fn)


def check_agent_spaces(
    agent_spaces: Dict[str, Space],
    space_type: str,
    check_box_space_fn: Callable[[spaces.Box], None],
):
    """A passive check of environment spaces that should not affect the environment."""
    assert all(isinstance(i, str) for i in agent_spaces)
    for i, space_i in agent_spaces.items():
        try:
            check_space(space_i, space_type, check_box_space_fn)
        except AssertionError as e:
            raise AssertionError(f"Invalid {space_type} space for agent '{i}'.") from e


check_state_space = partial(
    check_space,
    space_type="state",
    check_box_space_fn=_check_box_state_space,
)
check_agent_observation_spaces = partial(
    check_agent_spaces,
    space_type="observation",
    check_box_space_fn=_check_box_observation_space,
)
check_agent_action_spaces = partial(
    check_agent_spaces,
    space_type="action",
    check_box_space_fn=_check_box_action_space,
)


def check_state(state: M.StateType, model: M.POSGModel):
    """Check state is valid.

    Arguments
    ---------
    state: the state to check
    model: the environment model

    """
    if model.state_space is not None:
        assert model.state_space.contains(state)


def check_obs(obs, observation_space: spaces.Space, method_name: str):
    """Check the observation returned by the environment correspond to the declared one.

    Arguments
    ---------
    obs: The observation to check
    observation_space: The observation space of the observation
    method_name: The method name that generated the observation

    """
    pre = f"The obs returned by the `{method_name}()` method"
    if isinstance(observation_space, spaces.Discrete):
        if not isinstance(obs, (np.int64, int)):
            logger.warn(f"{pre} should be an int or np.int64, actual type: {type(obs)}")
    elif isinstance(observation_space, spaces.Box):
        if observation_space.shape != ():
            if not isinstance(obs, np.ndarray):
                logger.warn(
                    f"{pre} was expecting a numpy array, actual type: {type(obs)}"
                )
            elif obs.dtype != observation_space.dtype:
                logger.warn(
                    f"{pre} was expecting numpy array dtype to be "
                    f"{observation_space.dtype}, actual type: {obs.dtype}"
                )
    elif isinstance(observation_space, (spaces.MultiBinary, spaces.MultiDiscrete)):
        if not isinstance(obs, np.ndarray):
            logger.warn(f"{pre} was expecting a numpy array, actual type: {type(obs)}")
    elif isinstance(observation_space, spaces.Tuple):
        if not isinstance(obs, tuple):
            logger.warn(f"{pre} was expecting a tuple, actual type: {type(obs)}")
        assert len(obs) == len(observation_space.spaces), (
            f"{pre} length is not same as the observation space length, obs length: "
            f"{len(obs)}, space length: {len(observation_space.spaces)}"
        )
        for sub_obs, sub_space in zip(obs, observation_space.spaces):
            check_obs(sub_obs, sub_space, method_name)
    elif isinstance(observation_space, spaces.Dict):
        assert isinstance(obs, dict), f"{pre} must be a dict, actual type: {type(obs)}"
        assert obs.keys() == observation_space.spaces.keys(), (
            f"{pre} observation keys is not same as the observation space keys, obs "
            f"keys: {list(obs.keys())}, space keys: "
            f"{list(observation_space.spaces.keys())}"
        )
        for space_key in observation_space.spaces:
            check_obs(obs[space_key], observation_space[space_key], method_name)

    try:
        if obs not in observation_space:
            logger.warn(f"{pre} is not within the observation space.")
    except Exception as e:
        logger.warn(f"{pre} is not within the observation space with exception: {e}")


def check_agent_obs(
    obs: Dict[str, M.ObsType],
    observation_spaces: Dict[str, Space],
    method_name: str,
):
    """Check that each agent's observation returned by the environment is valid.

    Arguments
    ---------
    obs: The observation for each agent to check
    observation_spaces: The observation spaces for each agent's observation
    method_name: The method name that generated the observation

    """
    for i, obs_i in obs.items():
        try:
            check_obs(obs_i, observation_spaces[i], method_name)
        except AssertionError as e:
            raise AssertionError("Invalid observation for agent `{i}`.") from e


def check_reset_obs(obs: Dict[str, M.ObsType], model: M.POSGModel):
    """Check agent observations returned by the environment `reset()` method are valid.

    Arguments
    ---------
    obs: The observation for each agent to check
    model: The environment model

    """
    assert isinstance(obs, dict), (
        "Expected observation from `env.reset()` to be a dictionary, mapping agent IDs "
        " to agent obs."
    )
    for i, o_i in obs.items():
        assert (
            i in model.possible_agents
        ), f"Invalid agent ID `{i}`. Possible IDs are {model.possible_agents}."

    check_agent_obs(obs, model.observation_spaces, "reset")


def env_reset_passive_checker(env, **kwargs):
    """A passive check of the `Env.reset` function.

    Investigates the returning reset information and returning the data unchanged.

    """
    signature = inspect.signature(env.reset)
    if "seed" not in signature.parameters and "kwargs" not in signature.parameters:
        logger.warn(
            "posggym requires that `Env.reset` can be passed a `seed` for resetting the"
            " environment random number generator."
        )
    else:
        seed_param = signature.parameters.get("seed")
        # Check the default value is None
        if seed_param is not None and seed_param.default is not None:
            logger.warn(
                "The default seed argument in `Env.reset` should be `None`, otherwise "
                "the environment will by default always be deterministic. "
                f"Actual default: {seed_param}"
            )

    if "options" not in signature.parameters and "kwargs" not in signature.parameters:
        logger.warn(
            "posggym requires that `Env.reset` can be passed `options` to allow the "
            "environment initialisation to be passed additional information."
        )

    # Checks the result of env.reset with kwargs
    result = env.reset(**kwargs)

    if not isinstance(result, tuple):
        logger.warn(
            "The result returned by `env.reset()` was not a tuple of the form "
            "`(obs, info)`, where `obs` is a observation and `info` is a dictionary "
            f"containing additional information. Actual type: `{type(result)}`"
        )
    elif len(result) != 2:
        logger.warn(
            "The result returned by `env.reset()` should be `(obs, info)` by default, "
            "where `obs` is a observation and `info` is a dictionary containing "
            "additional information."
        )
    else:
        obs, info = result
        check_reset_obs(obs, env.model)
        assert isinstance(info, dict), (
            "The second element returned by `env.reset()` was not a dictionary, "
            f"actual type: {type(info)}"
        )
    return result


def _check_agent_dict(
    agent_dict,
    possible_agents: Sequence[str],
    dict_type: str,
    expected_agents: Optional[Sequence[str]] = None,
):
    assert isinstance(agent_dict, dict), (
        f"Agent {dict_type} dictionary  must be a dictionary mapping agentID to values."
        f"Actual type: {type(agent_dict)}."
    )
    for i in agent_dict:
        assert i in possible_agents, (
            f"Agent {dict_type} dictionary must only contain valid agent IDs: "
            f"invalid ID `{i}`."
        )
    if expected_agents is not None:
        for i in expected_agents:
            assert (
                i in agent_dict
            ), f"Expected agent ID `{i}` missing from {dict_type} dictionary."


def model_step_passive_checker(
    model: M.POSGModel, state: M.StateType, actions: Dict[str, M.ActType]
):
    """A passive check for the model step.

    Investigating the returning data then returning the data unchanged.
    """
    # List of agents still active in the state and which we expect to get data for
    step_agents = model.get_agents(state)
    for i in step_agents:
        assert (
            i in model.possible_agents
        ), f"Agent ID not in list of possible IDs. Invalid agent ID '{i}'. "

    # We don't check the actions as out-of-bounds values are allowed in some
    # environments
    result = model.step(state, actions)
    assert isinstance(result, M.JointTimestep), (
        "Expects model.step result to be a `posggym.model.JointTimestep`, "
        f"actual type: {type(result)}"
    )
    next_state, obs, reward, terminated, truncated, done, info = result

    _check_agent_dict(obs, model.possible_agents, "observation", step_agents)
    _check_agent_dict(reward, model.possible_agents, "reward", step_agents)
    _check_agent_dict(terminated, model.possible_agents, "terminated", step_agents)
    _check_agent_dict(truncated, model.possible_agents, "truncated", step_agents)
    # Less strict on checking there are entries for all agents in info values
    # as these are not functionally critical and are more for record keeping
    _check_agent_dict(info, model.possible_agents, "info")

    check_state(next_state, model)
    check_agent_obs(obs, model.observation_spaces, "step")

    if not all(isinstance(t_i, (bool, np.bool_)) for t_i in terminated.values()):
        logger.warn(
            "Expects `terminated` signal to be a boolean for every agent, "
            f"actual types: {[type(t_i) for t_i in terminated.values()]}."
        )
    if not all(isinstance(t_i, (bool, np.bool_)) for t_i in truncated.values()):
        logger.warn(
            "Expects `truncated` signal to be a boolean for every agent, "
            f"actual types: {[type(t_i) for t_i in truncated.values()]}."
        )

    if not isinstance(done, (bool, np.bool_)):
        logger.warn(
            "Expects `done` signal returned by `step()` to be a boolean, "
            f"actual type: {type(info)}"
        )

    if not (
        all(
            np.issubdtype(type(r_i), np.integer)
            or np.issubdtype(type(r_i), np.floating)
            for r_i in reward.values()
        )
    ):
        logger.warn(
            "The reward returned for each agent by `step()` must be a float, int, "
            "np.integer or np.floating, "
            f"actual types: {[type(r_i) for r_i in reward.values()]}."
        )
    else:
        for i, r_i in reward.items():
            if np.isnan(r_i):  # type: ignore
                logger.warn(f"The reward for agent `{i}` is a NaN value.")
            if np.isinf(r_i):  # type: ignore
                logger.warn(f"The reward for agent `{i}` is an inf value.")

    return result


def env_step_passive_checker(env: posggym.Env, actions: Dict[str, M.ActType]):
    """A passive check for the environment step.

    Investigating the returning data then returning the data unchanged.
    """
    # List of agents still active in the state and which we expect to get data for
    step_agents = env.agents
    for i in step_agents:
        assert (
            i in env.possible_agents
        ), f"Agent ID not in list of possible IDs. Invalid agent ID '{i}'. "

    # We don't check the actions as out-of-bounds values are allowed in some
    # environments
    result = env.step(actions)
    assert isinstance(
        result, tuple
    ), f"Expects step result to be a tuple, actual type: {type(result)}"
    if len(result) == 6:
        obs, reward, terminated, truncated, done, info = result

        _check_agent_dict(obs, env.possible_agents, "observation", step_agents)
        _check_agent_dict(reward, env.possible_agents, "reward", step_agents)
        _check_agent_dict(terminated, env.possible_agents, "terminated", step_agents)
        _check_agent_dict(truncated, env.possible_agents, "truncated", step_agents)
        # Less strict on checking entries for all agents in the outcomes
        # and info values as these are not functionally critical and are more for
        # record keeping
        _check_agent_dict(info, env.possible_agents, "info")

        if not all(isinstance(t_i, (bool, np.bool_)) for t_i in terminated.values()):
            logger.warn(
                "Expects `terminated` signal to be a boolean for every agent, "
                f"actual types: {[type(t_i) for t_i in terminated.values()]}."
            )
        if not all(isinstance(t_i, (bool, np.bool_)) for t_i in truncated.values()):
            logger.warn(
                "Expects `truncated` signal to be a boolean for every agent, "
                f"actual types: {[type(t_i) for t_i in truncated.values()]}."
            )

        if not isinstance(done, (bool, np.bool_)):
            logger.warn(
                "Expects `done` signal returned by `step()` to be a boolean, "
                f"actual type: {type(info)}"
            )
    else:
        raise error.Error(
            "Expected `Env.step` to return a six element tuple, actual number of "
            f"elements returned: {len(result)}."
        )

    check_agent_obs(obs, env.observation_spaces, "step")

    if not (
        all(
            np.issubdtype(type(r_i), np.integer)
            or np.issubdtype(type(r_i), np.floating)
            for r_i in reward.values()
        )
    ):
        logger.warn(
            "The reward returned for each agent by `step()` must be a float, int, "
            "np.integer or np.floating, "
            f"actual types: {[type(r_i) for r_i in reward.values()]}."
        )
    else:
        for i, r_i in reward.items():
            if np.isnan(r_i):  # type: ignore
                logger.warn(f"The reward for agent `{i}` is a NaN value.")
            if np.isinf(r_i):  # type: ignore
                logger.warn(f"The reward for agent `{i}` is an inf value.")

    return result


def _check_render_return(render_mode, render_return):
    """Produces warning if `render_return` doesn't match `render_mode`."""
    if render_mode == "human":
        if render_return is not None:
            logger.warn(
                f"Human rendering should return `None`, got {type(render_return)}"
            )
    elif render_mode == "rgb_array":
        if not isinstance(render_return, np.ndarray):
            logger.warn(
                "RGB-array rendering should return a numpy array, got "
                f"{type(render_return)}"
            )
        else:
            if render_return.dtype != np.uint8:
                logger.warn(
                    "RGB-array rendering should return a numpy array with dtype "
                    f"uint8, got {render_return.dtype}"
                )
            if render_return.ndim != 3:
                logger.warn(
                    "RGB-array rendering should return a numpy array with three axes, "
                    f"got {render_return.ndim}"
                )
            if render_return.ndim == 3 and render_return.shape[2] != 3:
                logger.warn(
                    "RGB-array rendering should return a numpy array in which the "
                    f"last axis has three dimensions, got {render_return.shape[2]}"
                )
    elif render_mode == "depth_array":
        if not isinstance(render_return, np.ndarray):
            logger.warn(
                "Depth-array rendering should return a numpy array, got "
                f"{type(render_return)}"
            )
        elif render_return.ndim != 2:
            logger.warn(
                "Depth-array rendering should return a numpy array with two axes, "
                f"got {render_return.ndim}"
            )
    elif render_mode in ["ansi", "ascii"]:
        if not isinstance(render_return, str):
            logger.warn(
                "ANSI/ASCII rendering should produce a string, got "
                f"{type(render_return)}"
            )
    elif render_mode.endswith("_list"):
        if not isinstance(render_return, list):
            logger.warn(
                "Render mode `{render_mode}` should produce a list, got "
                f"{type(render_return)}"
            )
        else:
            base_render_mode = render_mode[: -len("_list")]
            for item in render_return:
                _check_render_return(
                    base_render_mode, item
                )  # Check that each item of the list matches the base render mode
    elif render_mode.endswith("_dict"):
        # check posggym specific render modes, namely dict renders which
        # return mapping from agentID to agent specific render.
        if not isinstance(render_return, dict):
            logger.warn(
                f"Render mode `{render_mode}` should produce a dict, got "
                f"{type(render_return)}"
            )
        else:
            # Check that each item of the dict matches the base render mode
            # Doesn't check dict keys, i.e. agent IDs, are valid
            base_render_mode = render_mode[: -len("_dict")]
            for item in render_return:
                _check_render_return(base_render_mode, item)


def env_render_passive_checker(env: posggym.Env):
    """A passive check of the `Env.render`.

    Checks that render modes/fps in the metadata of the environment are declared.

    This function is exactly the same as
    `gymnasium.utils.passive_env_check.env_render_passive_checker`, we copy it here to
    ensure the updated `_check_render_return` function is used.

    """
    render_modes = env.metadata.get("render_modes")
    if render_modes is None:
        logger.warn(
            "No render modes were declared in the environment "
            "(env.metadata['render_modes'] is None or not defined), you may have "
            "trouble when calling `.render()`."
        )
    else:
        if not isinstance(render_modes, (list, tuple)):
            logger.warn(
                "Expects the render_modes to be a sequence (i.e. list, tuple), "
                f"actual type: {type(render_modes)}"
            )
        elif not all(isinstance(mode, str) for mode in render_modes):
            logger.warn(
                "Expects all render modes to be strings, "
                f"actual types: {[type(mode) for mode in render_modes]}"
            )

        render_fps = env.metadata.get("render_fps")
        # We only require `render_fps` if rendering is actually implemented
        if len(render_modes) > 0:
            if render_fps is None:
                logger.warn(
                    "No render fps was declared in the environment "
                    "(env.metadata['render_fps'] is None or not defined), rendering "
                    "may occur at inconsistent fps."
                )
            else:
                if not (
                    np.issubdtype(type(render_fps), np.integer)
                    or np.issubdtype(type(render_fps), np.floating)
                ):
                    logger.warn(
                        "Expects the `env.metadata['render_fps']` to be an integer "
                        f"or a float, actual type: {type(render_fps)}"
                    )
                else:
                    assert render_fps > 0, (
                        "Expects the `env.metadata['render_fps']` to be greater than "
                        f"zero, actual value: {render_fps}"
                    )

        # env.render is now an attribute with default None
        if len(render_modes) == 0:
            assert env.render_mode is None, (
                "With no render_modes, expects the Env.render_mode to be None, "
                f"actual value: {env.render_mode}"
            )
        else:
            assert env.render_mode is None or env.render_mode in render_modes, (
                "The environment was initialized successfully however with an "
                "unsupported render mode. "
                f"Render mode: {env.render_mode}, modes: {render_modes}"
            )

    result = env.render()  # type: ignore
    if env.render_mode is not None:
        _check_render_return(env.render_mode, result)

    return result



"""Root '__init__' of the posggym package."""
# isort: skip_file
# Need to import model and core before other modules
from posggym.model import POSGFullModel, POSGModel
from posggym.core import (
    ActionWrapper,
    Env,
    DefaultEnv,
    ObservationWrapper,
    RewardWrapper,
    Wrapper,
)
from posggym import envs, error, logger, utils, vector, wrappers
from posggym.envs import make, pprint_registry, register, registry, spec


__all__ = [
    # core classes
    "Env",
    "DefaultEnv",
    "Wrapper",
    "ObservationWrapper",
    "ActionWrapper",
    "RewardWrapper",
    "POSGModel",
    "POSGFullModel",
    # registration
    "make",
    "pprint_registry",
    "register",
    "registry",
    "spec",
    # module folders
    "envs",
    "utils",
    "vector",
    "wrappers",
    "error",
    "logger",
]
__version__ = "0.5.0"



"""Wrapper to discretize continuous actions."""
from typing import Dict, Sequence, Union, cast

import numpy as np
from gymnasium import spaces

from posggym import ActionWrapper, Env


class DiscretizeActions(ActionWrapper):
    """Action wrapper that discretizes continuous action space.

    The base environment :attr:`env` must have an action space of type
    :class:`spaces.Box` for each agent. If :attr:`flatten` is True and original action
    space is multi-dimensional with :attr:`ndim` dimensions, then will create
    discretized space with :attr:`num_actions ** ndim` actions.

    Arguments
    ---------
    env : posggym.Env
        The environment to apply the wrapper
    num_actions : int
        The number of actions to discretize space into. For multi-dimensional
        continuous spaces each dimension will be discretized into this many actions.
    flatten : bool
       Whether to flatten action space into one-dimensional discrete space, or keep
       number of dimensions of the original action space.

    """

    def __init__(self, env: Env, num_actions: int, flatten: bool = False):
        super().__init__(env)
        assert all(
            isinstance(act_space, spaces.Box)
            for act_space in self.action_spaces.values()
        )

        self.num_actions = num_actions
        self.flatten = flatten

        box_action_spaces = cast(Dict[str, spaces.Box], self.action_spaces)
        self._unflat_space: Dict[str, spaces.MultiDiscrete] = {}
        if self.flatten:
            self._unflat_space = {
                i: self.discretize_action_space(  # type: ignore
                    act_space, num_actions=num_actions, flatten=False
                )
                for i, act_space in box_action_spaces.items()
            }

        self._action_spaces = {
            i: self.discretize_action_space(
                act_space, num_actions=num_actions, flatten=flatten
            )
            for i, act_space in box_action_spaces.items()
        }

    def discretize_action_space(
        self, action_space: spaces.Box, num_actions: int, flatten: bool = False
    ) -> Union[spaces.MultiDiscrete, spaces.Discrete]:
        assert isinstance(action_space, spaces.Box), "Action space must be a Box"
        assert (
            len(action_space.shape) > 0
        ), "Action space must be at least 1-dimensional"
        assert (
            action_space.is_bounded()
        ), "Action space must be bounded to discretize it!"

        num_dims = action_space.shape[0]
        md_space = spaces.MultiDiscrete([num_actions] * num_dims)
        if flatten:
            # Compute the product of the individual action space sizes
            n = np.prod(md_space.nvec)
            return spaces.Discrete(n)
        return md_space

    def actions(self, actions):
        if self.flatten:
            # first unflatten
            actions = {
                i: self.multidiscretize_discrete_action(act_i, self._unflat_space[i])
                for i, act_i in actions.items()
            }
        return {
            i: self.undiscretize_action(act_i, self.model.action_spaces[i])
            for i, act_i in actions.items()
        }

    def undiscretize_action(
        self,
        discrete_action: Union[int, Sequence[int], np.ndarray],
        action_space: spaces.Box,
    ) -> np.ndarray:
        if isinstance(discrete_action, int):
            discrete_action = [discrete_action]
        assert (
            len(discrete_action) == action_space.shape[0]
        ), "Discrete action must have the same number of dimensions as the action space"
        assert all(0 <= a < self.num_actions for a in discrete_action), (
            f"Discrete action must be between 0 and `num_actions`={self.num_actions}. "
            f"{discrete_action} invalid."
        )
        # Compute the bin size for each dimension
        bin_sizes = (action_space.high - action_space.low) / (self.num_actions - 1)
        # Convert each discrete action to a continuous value
        continuous_action = (
            action_space.low
            + np.array(discrete_action, dtype=action_space.dtype) * bin_sizes
        )
        return continuous_action

    def multidiscretize_discrete_action(
        self, discrete_action: int, multidiscrete_space: spaces.MultiDiscrete
    ):
        # Compute the individual action space sizes
        nvec = multidiscrete_space.nvec
        # Create an empty array to store the multidiscrete action
        action = np.zeros_like(nvec)
        # Compute the Cartesian product index for each action space
        for i in range(len(nvec)):
            prod = int(np.prod(nvec[i + 1 :]))
            action[i] = int(discrete_action // prod) % nvec[i]
        return action



"""The Cooperative Reaching Grid World Environment."""
from itertools import product
from os import path
from typing import Dict, List, Optional, Tuple

from gymnasium import spaces

import posggym.model as M
from posggym import logger
from posggym.core import DefaultEnv
from posggym.envs.grid_world.core import Coord, Direction, Grid
from posggym.utils import seeding


# State = (coord_0, coord_1)
CRState = Tuple[Coord, Coord]

# The actions
CRAction = int
DO_NOTHING = 0
UP = 1
DOWN = 2
LEFT = 3
RIGHT = 4

ACTIONS = [DO_NOTHING, UP, DOWN, LEFT, RIGHT]
ACTIONS_STR = ["0", "U", "D", "L", "R"]
ACTION_TO_DIR = [None, Direction.NORTH, Direction.SOUTH, Direction.WEST, Direction.EAST]

# Obs = (ego_coord, other_coord)
CRObs = Tuple[Coord, Coord]


class CooperativeReachingEnv(DefaultEnv[CRState, CRObs, CRAction]):
    """The Cooperative Reaching Grid World Environment.

    A cooperative 2D grid world problem where two agents must coordinate to go to the
    same goal. This environment tests an agent's ability to coordinate with another
    agent.

    Possible Agents
    ---------------
    The environment supports two agents, with both agents always beginning and ending
    each episode at the same time.

    State Space
    -----------
    Each state is made up of the the `(x, y)` coordinate of each agent. For the
    coordinate `x`=column, `y`=row, with the origin (0, 0) at the top-left square of
    the grid.

    Action Space
    ------------
    Each agent has 5 actions: `DO_NOTHING=0`, `UP=1`, `DOWN=2`, `LEFT=3`, `RIGHT=4`

    Observation Space
    -----------------
    Each agent observes their `(x, y)` coordinate, as well as the `(x, y)` coordinate
    of the other agent, as long as the other agent is within the observation range.
    If the other agent is outside the observation range then their observed coordinate
    will be `(size, size)` (i.e. outside of the grid).

    All together each agent's observation is tuple of the form:

        (ego coord, other coord)

    Rewards
    -------
    Both agents receive a reward when they simoultaneously reach the same goal cell.
    The reward they receive will depend on the value of the goal cell, which is
    determined by the scenario. For all other steps the agents receive a reward of
    `0.0`.

    Dynamics
    --------
    Actions are deterministic and consist of moving to the adjacent cell in each of the
    four cardinal directions. If an agent attempts to move out of bounds of the grid
    then they remain in their current cell. Agents can occupy the same cell at the same
    time.

    Starting State
    --------------
    Agents start from random locations in the middle of the grid.

    Episodes End
    ------------
    Episodes end when both agents simoultaneously reach the same goal cell. By default
    a `max_episode_steps` is also set. The default value is `50` steps, but this may
    need to be adjusted when using larger grids (this can be done by manually specifying
    a value for `max_episode_steps` when creating the environment with `posggym.make`).

    Arguments
    ---------

    - `size` - the size (width and height) of grid.
    - `num_goals` - the number of goal cells in the grid.
    - `mode` - the mode of the environment, which determines the layout of goals in the
        grid as well as their values. The available modes are: ["square", "line",
        "original"]
    - `obs_distance` - the number of cells in each direction that each agent can
        observe. This determines how close agents need to be to each other to be able
        to observe each other's location. Setting this to be `2*size` will make the
        environment fully observable (default = `None` = `2*size`).

    Available variants
    ------------------
    The Cooperative Reaching environment comes with a number of benchmark grid layouts
    which can be passed as an argument to `posggym.make` using the `mode` argument.

    - `original` - the original grid layout from the paper. This is a grid with four
        goals, one in each corner of the grid. The goal values are: top-left = 1,
        top-right = 0.75, bottom-right = 1, bottom-left = 0.75. Note, that this mode
        only supports having four goals (`num_goals=4`).
    - `square` - goals are spaced out evenly along the border of the grid, starting
        from the top-left corner and moving clockwise. Supports any number of goals
        `1 <= num_goals <= (size-1)*4` and all goals have the same value of `1.0`.
    - `line` - goals are laid out in a line evenly along the middle column of the grid
        (or one of the middle columns if the grid has an even number of columns).
        Supports any number of goals `1 <= num_goals <= size` and all goals have the
        same value of `1.0`.

    The following table are some standard benchmark layouts that have been used in
    papers or are similar to those studied in paper:

    | Name              | `size` | `num_goals` | `mode`   |
    |-------------------|--------|------------ | -------- |
    | `original_5`      | 5      | 4           | original |
    | `original_10`     | 10     | 4           | original |
    | `square_5_n4`     | 5      | 4           | square   |
    | `square_10_n4`    | 10     | 4           | square   |
    | `square_10_n8`    | 10     | 8           | square   |
    | `line_5_n3`       | 5      | 3           | line     |
    | `line_7_n4`       | 7      | 4           | line     |
    | `line_11_n6`      | 11     | 6           | line     |

    Note, "Name" here is just provided to give a label for each layout. To use one of
    these layouts the user must specify each argument.

    For example to use the Cooperative Reaching environment with the `square_5_n5`
    benchmark layout, you would use:

    ```python
    import posggym
    env = posggym.make('CooperativeReaching-v0', size=5, num_goals=4, mode="square")
    ```

    Version History
    ---------------
    - `v0`: Initial version

    References
    ----------
    - Arrasy Rahman, Elliot Fosong, Ignacio Carlucho, and Stefano V. Albrecht. 2023.
      Generating Teammates for Training Robust Ad Hoc Teamwork Agents via Best-Response
      Diversity. Transactions on Machine Learning Research.
    """

    metadata = {
        "render_modes": ["human", "ansi", "rgb_array", "rgb_array_dict"],
        "render_fps": 15,
    }

    def __init__(
        self,
        size: int = 5,
        num_goals: int = 4,
        mode: str = "original",
        obs_distance: Optional[int] = None,
        render_mode: Optional[str] = None,
    ):
        super().__init__(
            CooperativeReachingModel(size, num_goals, mode, obs_distance),
            render_mode=render_mode,
        )
        self.renderer = None
        self._agent_imgs = None

    def render(self):
        if self.render_mode is None:
            assert self.spec is not None
            logger.warn(
                "You are calling render method without specifying any render mode. "
                "You can specify the render_mode at initialization, "
                f'e.g. posggym.make("{self.spec.id}", render_mode="rgb_array")'
            )
            return
        if self.render_mode == "ansi":
            return self._render_ansi()
        return self._render_img()

    def _render_ansi(self):
        model: CooperativeReachingModel = self.model  # type: ignore
        grid_str = model.grid.get_ascii_repr(self._state)

        output = [
            f"Step: {self._step_num}",
            grid_str,
        ]
        if self._last_actions is not None:
            action_str = ", ".join(
                [ACTIONS_STR[a] for a in self._last_actions.values()]
            )
            output.insert(1, f"Actions: <{action_str}>")
            output.append(f"Rewards: <{self._last_rewards}>")

        return "\n".join(output) + "\n"

    def _render_img(self):
        model: CooperativeReachingModel = self.model  # type: ignore

        import posggym.envs.grid_world.render as render_lib

        if self.renderer is None:
            self.renderer = render_lib.GWRenderer(
                self.render_mode,
                model.grid,
                render_fps=self.metadata["render_fps"],
                env_name="CooperativeReaching",
            )

        if self._agent_imgs is None:
            img_path = path.join(path.dirname(__file__), "img", "robot.png")
            agent_img = render_lib.load_img_file(img_path, self.renderer.cell_size)
            self._agent_imgs = {
                i: render_lib.GWImage((0, 0), self.renderer.cell_size, agent_img)
                for i in self.possible_agents
            }

        observed_coords = []
        if model.obs_distance < model.size * 2:
            for coord in self._state:
                observed_coords.extend(model.get_obs_coords(coord))

        render_objects = []
        # Add goal locations. Color is green but shaded based on goal value
        for coord, value in model.goals.items():
            render_objects.append(
                render_lib.GWRectangle(
                    coord,
                    self.renderer.cell_size,
                    color=(
                        int(255 * (1 - value)),
                        int(255 * value),
                        int(255 * (1 - value)),
                        int(255 * value),
                    ),
                )
            )

        # Add agents
        for i, coord in enumerate(self._state):
            agent_obj = self._agent_imgs[str(i)]
            agent_obj.coord = coord
            render_objects.append(agent_obj)

        agent_coords_and_dirs = {
            str(i): (coord, Direction.NORTH) for i, coord in enumerate(self._state)
        }

        if self.render_mode in ("human", "rgb_array"):
            return self.renderer.render(render_objects, observed_coords)
        return self.renderer.render_agents(
            render_objects,
            agent_coords_and_dirs,
            agent_obs_dims=model.obs_distance,
            observed_coords=observed_coords,
        )

    def close(self) -> None:
        if self.renderer is not None:
            self.renderer.close()
            self.renderer = None


class CooperativeReachingModel(M.POSGModel[CRState, CRObs, CRAction]):
    """Driving Problem Model.

    Parameters
    ----------
    size : int
        the size (width and height) of grid.
    num_goals : int
        the number of goals in the grid.
    mode : str
        the mode of the environment, which determines the layout of goals in the
        grid as well as their values. The available modes are: ["square", "line",
        "original"]
    obs_distance : int or None
        the number of cells in each direction that each agent can observe. This
        determines how close agents need to be to each other to be able to observe each
        other's location. Setting this to be `2*size` will make the environment fully
        observable. If `None` then will be set to `2*size`.

    """

    NUM_AGENTS = 2

    MODES = ["square", "line", "original"]

    def __init__(
        self,
        size: int,
        num_goals: int,
        mode: str,
        obs_distance: Optional[int],
    ):
        assert size >= 3, "Grid size must be at least 3"
        assert num_goals >= 1, "Must have at least one goal"
        assert mode in self.MODES, f"Mode must be one of {self.MODES}"
        if obs_distance is None:
            obs_distance = 2 * size
        assert obs_distance >= 0, "Observation distance must be non-negative"

        self.size = size
        self.num_goals = num_goals
        self.mode = mode
        self.obs_distance = obs_distance

        if mode == "square":
            points = equid_points_square(num_goals, self.size)
            self.goals = {p: 1.0 for p in points}
        elif mode == "line":
            points = equid_points_line(num_goals, self.size)
            self.goals = {p: 1.0 for p in points}
        else:
            if num_goals != 4:
                logger.warn(
                    f"'original' mode only supports 4 goals, but got {num_goals}. "
                    "Continuing with 4 goals."
                )
            self.goals = {
                (0, 0): 1.0,
                (self.size - 1, self.size - 1): 1.0,
                (0, self.size - 1): 0.75,
                (self.size - 1, 0): 0.75,
            }

        self.grid = CooperativeReachingGrid(self.size, list(self.goals))
        # TODO review this. Do we want to restrict the start coords so agents start
        # within the observation range of each other?
        self.possible_start_coords = list(product(range(1, self.size - 1), repeat=2))

        def _coord_space(s):
            return spaces.Tuple((spaces.Discrete(s), spaces.Discrete(s)))

        self.possible_agents = tuple(str(i) for i in range(self.NUM_AGENTS))
        self.state_space = spaces.Tuple(
            tuple(_coord_space(size) for _ in range(len(self.possible_agents)))
        )
        self.action_spaces = {
            i: spaces.Discrete(len(ACTIONS)) for i in self.possible_agents
        }
        self.observation_spaces = {
            i: spaces.Tuple(
                (
                    _coord_space(size),
                    _coord_space(size + 1),  # +1 to account for non-observable agent
                )
            )
            for i in self.possible_agents
        }
        self.is_symmetric = True

    @property
    def reward_ranges(self) -> Dict[str, Tuple[float, float]]:
        max_goal_value = max(self.goals.values())
        return {i: (0.0, max_goal_value) for i in self.possible_agents}

    @property
    def rng(self) -> seeding.RNG:
        if self._rng is None:
            self._rng, _ = seeding.std_random()
        return self._rng

    def get_agents(self, state: CRState) -> List[str]:
        return list(self.possible_agents)

    def sample_initial_state(self) -> CRState:
        return tuple(
            self.rng.choice(self.possible_start_coords)
            for _ in range(len(self.possible_agents))
        )

    def sample_agent_initial_state(self, agent_id: str, obs: CRObs) -> CRState:
        assert (0 <= p < self.size for p in obs[1]), (
            "Invalid other agent coord for sampling agent-centric initial state. "
            "Other agent should be observable and within grid."
        )
        agent_idx = int(agent_id)
        return (obs[0], obs[1]) if agent_idx == 0 else (obs[1], obs[0])

    def sample_initial_obs(self, state: CRState) -> Dict[str, CRObs]:
        return self._get_obs(state)

    def step(
        self, state: CRState, actions: Dict[str, CRAction]
    ) -> M.JointTimestep[CRState, CRObs]:
        assert all(0 <= a_i < len(ACTIONS) for a_i in actions.values())
        next_state = self._get_next_state(state, actions)
        obs = self._get_obs(next_state)
        rewards = self._get_rewards(next_state)

        all_done = (
            all(p == next_state[0] for p in next_state) and next_state[0] in self.goals
        )
        terminated = {i: all_done for i in self.possible_agents}
        truncated = {i: False for i in self.possible_agents}

        info: Dict[str, Dict] = {i: {} for i in self.possible_agents}
        if all_done:
            for i in self.possible_agents:
                info[i]["outcome"] = M.Outcome.WIN

        return M.JointTimestep(
            next_state, obs, rewards, terminated, truncated, all_done, info
        )

    def _get_next_state(self, state: CRState, actions: Dict[str, CRAction]) -> CRState:
        next_state = list(state)
        for i, action_i in actions.items():
            if action_i == DO_NOTHING:
                continue
            idx = int(i)
            next_state[idx] = self.grid.get_next_coord(
                next_state[idx], ACTION_TO_DIR[action_i], ignore_blocks=False
            )
        return tuple(next_state)

    def _get_obs(self, state: CRState) -> Dict[str, CRObs]:
        obs: Dict[str, CRObs] = {}
        for i in self.possible_agents:
            idx = int(i)
            other_idx = (idx + 1) % 2
            state_i, state_j = state[idx], state[other_idx]

            if (
                abs(state_i[0] - state_j[0]) <= self.obs_distance
                and abs(state_i[1] - state_j[1]) <= self.obs_distance
            ):
                obs[i] = (state_i, state_j)
            else:
                obs[i] = (state_i, (self.size, self.size))
        return obs

    def get_obs_coords(self, origin: Coord) -> List[Coord]:
        """Get the list of coords observed from agent at origin."""
        obs_size = (2 * self.obs_distance) + 1
        obs_coords: List[Coord] = []
        for obs_col, obs_row in product(range(obs_size), repeat=2):
            grid_col = origin[0] + obs_col - self.obs_distance
            grid_row = origin[1] + obs_row - self.obs_distance
            if 0 <= grid_row < self.size and 0 <= grid_col < self.size:
                obs_coords.append((grid_col, grid_row))
        return obs_coords

    def _get_rewards(self, state: CRState) -> Dict[str, float]:
        all_done = all(p == state[0] for p in state) and state[0] in self.goals
        if all_done:
            goal_value = self.goals[state[0]]
            return {i: goal_value for i in self.possible_agents}
        return {i: 0.0 for i in self.possible_agents}


class CooperativeReachingGrid(Grid):
    """A grid for the Cooperative Reaching Problem."""

    def __init__(
        self,
        size: int,
        goal_coords: List[Coord],
    ):
        assert size >= 3, "Grid size must be at least 3"
        super().__init__(size, size, block_coords=set())
        self.size = size
        self.goal_coords = goal_coords
        self.shortest_paths = self.get_all_shortest_paths(goal_coords)

    def get_shortest_path_distance(self, coord: Coord, goal: Coord) -> int:
        """Get the shortest path distance from coord to goal."""
        return int(self.shortest_paths[goal][coord])

    def get_ascii_repr(self, agent_coords: Optional[CRState]) -> str:
        """Get ascii repr of grid."""
        grid_repr = []
        for row in range(self.height):
            row_repr = []
            for col in range(self.width):
                coord = (col, row)
                if coord in self.block_coords:
                    row_repr.append("#")
                elif coord in self.goal_coords:
                    row_repr.append("G")
                else:
                    row_repr.append(".")
            grid_repr.append(row_repr)

        if agent_coords is not None:
            for i, coord in enumerate(agent_coords):
                grid_repr[coord[0]][coord[1]] = str(i)

        return "\n".join([" ".join(r) for r in grid_repr])


def equid_points_square(n_points: int, grid_size: int) -> List[Tuple[int, int]]:
    """Return n_points equidistant points on square border of grid."""
    assert 0 < n_points <= (grid_size - 1) * 4
    perimeter_length = (grid_size - 1) * 4
    dp = perimeter_length / n_points
    perimeter_idxs = [int(0.5 * dp + i * dp) for i in range(n_points)]
    # need to shift each index so first point is at (0, 0)
    perimeter_idxs = [i - perimeter_idxs[0] for i in perimeter_idxs]
    points = []
    for i in perimeter_idxs:
        if i < grid_size - 1:
            points.append((i, 0))
        elif i < 2 * (grid_size - 1):
            points.append((grid_size - 1, i - (grid_size - 1)))
        elif i < 3 * (grid_size - 1):
            points.append((3 * (grid_size - 1) - i, grid_size - 1))
        else:
            points.append((0, 4 * (grid_size - 1) - i))
    return points


def equid_points_line(n_points: int, grid_size: int) -> List[Tuple[int, int]]:
    """Return n_points equidistant points on line in middle of grid."""
    assert 0 < n_points <= grid_size
    col = int(grid_size / 2)
    dr = grid_size / n_points
    points = [(col, int(0.5 * dr + i * dr)) for i in range(n_points)]
    return points



"""The Multi-Access Broadcast problem."""
import sys
from itertools import product
from typing import Dict, List, Optional, Tuple, Union

from gymnasium import spaces

import posggym.model as M
from posggym import logger
from posggym.core import DefaultEnv
from posggym.utils import seeding

MABCState = Tuple[int, ...]
EMPTY = 0
FULL = 1
NODE_STATES = [EMPTY, FULL]
NODE_STATE_STR = ["E", "F"]

MABCAction = int
SEND = 0
NOSEND = 1
ACTIONS = [SEND, NOSEND]
ACTION_STR = ["S", "NS"]

MABCObs = int
COLLISION = 0
NOCOLLISION = 1
OBS = [COLLISION, NOCOLLISION]
OBS_STR = ["C", "NC"]


class MABCEnv(DefaultEnv[MABCState, MABCObs, MABCAction]):
    """The Multi-Access Broadcast Channel Environment.

    A cooperative game involving control of a multi-access broadcast channel.
    In this problem, each agent controls a node in a network. Each node needs to
    broadcast messages to each other over a shared channel, with only one node able to
    broadcast at a time. If more than one node broadcasts at the same time then there
    is a collision and no message is broadcast. The nodes share the common goal of
    maximizing the throughput of the channel.

    Possible Agents
    ---------------
    The environment supports two or more agents, although the default version only
    supports two agents. All agents are always active in the environment.

    State Space
    -----------
    Each node has a message buffer that can store up to one message at a time.
    That is it can be either `EMPTY=0` or `FULL=1`.

    Action Space
    ------------
    At each timestep each agent can either `SEND=0` a message or not `NOSEND=1`.

    Observation Space
    -----------------
    At the end of each time step, each node receives a noisy observation of
    whether there was a `COLLISION=0` or `NOCOLLISION=1`.

    Each agent observes the true outcome with probability `obs_prob`, which is
    0.9 in the default version.

    Rewards
    -------
    Each agent receives a reward of `1` when a message is successfully
    broadcast and a reward of `0` otherwise.

    Dynamics
    --------
    If a node's buffer is EMPTY then at each step it will become full with
    probability `fill_probs[i]`,  otherwise it will remain empty (independent
    of the action performed). Where `i` is the agent ID. By default
    `fill_probs` for node 0 is 0.9, and for node 1 is 0.1 (as per the paper).

    If a node's buffer is `FULL` and the node does not send a message - i.e. uses the
    `NOSEND` action - then the buffer remains `FULL`. Otherwise - i.e. the agent chooses
    the `SEND` action - if no other nodes sends a message (there is no COLLISION) the
    buffer will be FULL with probability `fill_probs[i]`, otherwise it will be empty.
    If another message was sent at the same time by another node then there will be a
    COLLISION and the node's buffer remains `FULL`.

    Starting State
    --------------
    Each node buffer starts as `FULL` with probability `init_buffer_dist[i]` (which is
    `1.0` by default), otherwise the buffer starts as `EMPTY`.

    Episode End
    -----------
    By default episodes continue infinitely long. To set a step limit, specify
    `max_episode_steps` when initializing the environment with `posggym.make`.

    Arguments
    ---------

    - `num_nodes` - the number of nodes (i.e. agents) in the network (default=`2.0`)
    - `fill_probs` - the probability each nodes buffer is filled, should be a tuple with
        an entry for each node (default = `None` = `(0.9, 0.1)`)
    - `observation_prob` - the probability of correctly observing if there was a
        collision or not (default = `0.9`)
    - `init_buffer_dist` - the probability each node starts with a full buffer. Should
        be a tuple with an entry for each node (default = `None` = `(1.0, 1.0)`)

    Version History
    ---------------
    - `v0`: Initial version

    References
    ----------
    - Ooi, J. M., and Wornell, G. W. 1996. Decentralized control of a multiple
      access broadcast channel: Performance bounds. In Proceedings of the 35th
      Conference on Decision and Control, 293–298.
    - Hansen, Eric A., Daniel S. Bernstein, and Shlomo Zilberstein. “Dynamic
      Programming for Partially Observable Stochastic Games.” In Proceedings of
      the 19th National Conference on Artificial Intelligence, 709–715. AAAI’04.
      San Jose, California: AAAI Press, 2004.

    """

    metadata = {"render_modes": ["human", "ansi"], "render_fps": 4}

    def __init__(
        self,
        num_nodes: int = 2,
        fill_probs: Optional[Tuple[float, ...]] = None,
        observation_prob: float = 0.9,
        init_buffer_dist: Optional[Tuple[float, ...]] = None,
        render_mode: Optional[str] = None,
    ):
        super().__init__(
            MABCModel(num_nodes, fill_probs, observation_prob, init_buffer_dist),
            render_mode=render_mode,
        )

    def render(self):
        if self.render_mode is None:
            assert self.spec is not None
            logger.warn(
                "You are calling render method without specifying any render mode. "
                "You can specify the render_mode at initialization, "
                f'e.g. posggym.make("{self.spec.id}", render_mode="rgb_array")'
            )
            return

        assert self._last_obs is not None

        state_str = ", ".join([NODE_STATE_STR[s] for s in self._state])
        obs_str = ", ".join([OBS_STR[o] for o in self._last_obs.values()])
        output = [
            f"Step: {self._step_num}",
            f"State: <{state_str}>",
            f"Obs: <{obs_str}>",
        ]
        if self._last_actions is not None:
            action_str = ", ".join([ACTION_STR[a] for a in self._last_actions.values()])
            output.insert(1, f"Actions: <{action_str}>")

        if self._last_rewards is not None:
            output.append(f"Rewards: <{tuple(self._last_rewards.values())}>")

        output_str = "\n".join(output) + "\n"

        if self.render_mode == "human":
            sys.stdout.write(output_str)
        else:
            # ansi mode
            return output_str


class MABCModel(M.POSGFullModel[MABCState, MABCObs, MABCAction]):
    """POSG Model for the Multi-Access Broadcast Channel problem."""

    DEFAULT_FILL_PROBS = (0.9, 0.1)
    DEFAULT_OBS_PROB = 0.9
    DEFAULT_INIT_BUFFER_DIST = (1.0, 1.0)

    R_SEND = 1.0
    R_NO_SEND = 0.0

    def __init__(
        self,
        num_nodes: int = 2,
        fill_probs: Optional[Tuple[float, ...]] = None,
        observation_prob: float = 0.9,
        init_buffer_dist: Optional[Tuple[float, ...]] = None,
    ):
        assert num_nodes >= 2

        if fill_probs is None:
            fill_probs = self.DEFAULT_FILL_PROBS
        if init_buffer_dist is None:
            init_buffer_dist = self.DEFAULT_INIT_BUFFER_DIST

        assert len(fill_probs) == num_nodes
        assert all(0 <= x <= 1 for x in fill_probs)
        assert 0 <= observation_prob <= 1
        assert len(init_buffer_dist) == num_nodes
        assert all(0 <= x <= 1 for x in init_buffer_dist)

        self._fill_probs = fill_probs
        self._obs_prob = observation_prob
        self._init_buffer_dist = init_buffer_dist

        self.possible_agents = tuple(str(i) for i in range(num_nodes))
        self.state_space = spaces.Tuple(
            tuple(spaces.Discrete(len(NODE_STATES)) for _ in self.possible_agents)
        )
        self.action_spaces = {
            i: spaces.Discrete(len(ACTIONS)) for i in self.possible_agents
        }
        self.observation_spaces = {
            i: spaces.Discrete(len(OBS)) for i in self.possible_agents
        }
        self.is_symmetric = True

        # Spaces used internally
        self._state_space = list(
            product(*[list(NODE_STATES) for _ in self.possible_agents])
        )
        self._action_spaces = tuple([*ACTIONS] for _ in self.possible_agents)
        self._observation_spaces = tuple([*OBS] for _ in self.possible_agents)

        self._trans_map = self._construct_trans_func()
        self._rew_map = self._construct_rew_func()
        self._obs_map = self._construct_obs_func()

    @property
    def reward_ranges(self) -> Dict[str, Tuple[float, float]]:
        return {i: (self.R_NO_SEND, self.R_SEND) for i in self.possible_agents}

    @property
    def rng(self) -> seeding.RNG:
        if self._rng is None:
            self._rng, seed = seeding.std_random()
        return self._rng

    def get_agents(self, state: MABCState) -> List[str]:
        return list(self.possible_agents)

    def sample_initial_state(self) -> MABCState:
        node_states = []
        for i in range(len(self.possible_agents)):
            if self.rng.random() <= self._init_buffer_dist[i]:
                node_states.append(FULL)
            else:
                node_states.append(EMPTY)
        return tuple(node_states)

    def sample_initial_obs(self, state: MABCState) -> Dict[str, MABCObs]:
        return {i: NOCOLLISION for i in self.possible_agents}

    def step(
        self, state: MABCState, actions: Dict[str, MABCAction]
    ) -> M.JointTimestep[MABCState, MABCObs]:
        assert all(a_i in ACTIONS for a_i in actions.values())
        next_state = self._sample_next_state(state, actions)
        obs = self._sample_obs(actions)
        agent_reward = float(self._message_sent(state, actions)) * self.R_SEND
        rewards = {i: agent_reward for i in self.possible_agents}
        terminated = {i: False for i in self.possible_agents}
        truncated = {i: False for i in self.possible_agents}
        all_done = False
        info: Dict[str, Dict] = {i: {} for i in self.possible_agents}

        return M.JointTimestep(
            next_state, obs, rewards, terminated, truncated, all_done, info
        )

    def _sample_next_state(
        self, state: MABCState, actions: Dict[str, MABCAction]
    ) -> MABCState:
        next_node_states = list(state)
        for i, a_i in actions.items():
            idx = int(i)
            if a_i == SEND:
                # buffer emptied even if there is a collision
                next_node_states[idx] = EMPTY
            if self.rng.random() <= self._fill_probs[idx]:
                next_node_states[idx] = FULL
        return tuple(next_node_states)

    def _sample_obs(self, actions: Dict[str, MABCAction]) -> Dict[str, MABCObs]:
        senders = sum(int(a_i == SEND) for a_i in actions.values())
        if senders > 1:
            correct_obs = COLLISION
            wrong_obs = NOCOLLISION
        else:
            correct_obs = NOCOLLISION
            wrong_obs = COLLISION

        obs = {}
        for i in self.possible_agents:
            if self.rng.random() <= self._obs_prob:
                obs[i] = correct_obs
            else:
                obs[i] = wrong_obs
        return obs

    def get_initial_belief(self) -> Dict[MABCState, float]:
        b_map: Dict[MABCState, float] = {}
        s_prob_sum = 0.0
        for s in self._state_space:
            s_prob = 1.0
            for i in range(len(self.possible_agents)):
                if s[i] == FULL:
                    s_prob *= self._init_buffer_dist[i]
                else:
                    s_prob *= 1 - self._init_buffer_dist[i]
            b_map[s] = s_prob
            s_prob_sum += s_prob

        for s in self._state_space:
            b_map[s] /= s_prob_sum

        return b_map

    def transition_fn(
        self,
        state: MABCState,
        actions: Dict[str, MABCAction],
        next_state: MABCState,
    ) -> float:
        action_tuple = tuple(actions[i] for i in self.possible_agents)
        return self._trans_map[(state, action_tuple, next_state)]

    def _construct_trans_func(self) -> Dict:
        trans_map = {}
        agent_ids = [int(i) for i in self.possible_agents]
        for s, a, s_next in product(
            self._state_space, product(*self._action_spaces), self._state_space
        ):
            trans_prob = 1.0
            for i in agent_ids:
                if a[i] == NOSEND and s[i] == FULL and s_next[i] != FULL:
                    trans_prob *= 0.0
                elif s_next[i] == FULL:
                    trans_prob *= self._fill_probs[i]
                else:
                    trans_prob *= 1 - self._fill_probs[i]

            trans_map[(s, a, s_next)] = trans_prob
        return trans_map

    def observation_fn(
        self,
        obs: Dict[str, MABCObs],
        next_state: MABCState,
        actions: Dict[str, MABCAction],
    ) -> float:
        obs_tuple = tuple(obs[i] for i in self.possible_agents)
        action_tuple = tuple(actions[i] for i in self.possible_agents)
        return self._obs_map[(next_state, action_tuple, obs_tuple)]

    def _construct_obs_func(self) -> Dict:
        obs_map = {}
        agent_ids = [int(i) for i in self.possible_agents]
        for s_next, a, o in product(
            self._state_space,
            product(*self._action_spaces),
            product(*self._observation_spaces),
        ):
            senders = sum(int(a_i == SEND) for a_i in a)
            correct_obs = COLLISION if senders > 1 else NOCOLLISION

            o_prob = 1.0
            for i in agent_ids:
                if o[i] == correct_obs:
                    o_prob *= self._obs_prob
                else:
                    o_prob *= 1 - self._obs_prob
            obs_map[(s_next, a, o)] = o_prob
        return obs_map

    def reward_fn(
        self, state: MABCState, actions: Dict[str, MABCAction]
    ) -> Dict[str, float]:
        action_tuple = tuple(actions[i] for i in self.possible_agents)
        return self._rew_map[(state, action_tuple)]

    def _construct_rew_func(self) -> Dict:
        rew_map = {}
        joint_actions_space = product(*self._action_spaces)
        for s, a in product(self._state_space, joint_actions_space):
            reward = float(self._message_sent(s, a)) * self.R_SEND
            rew_map[(s, a)] = tuple(reward for _ in self.possible_agents)
        return rew_map

    def _message_sent(
        self,
        state: MABCState,
        actions: Union[Dict[str, MABCAction], Tuple[MABCAction, ...]],
    ) -> bool:
        if isinstance(actions, dict):
            actions = tuple(actions[i] for i in self.possible_agents)

        senders = sum(int(a_i == SEND) for a_i in actions)
        if senders != 1:
            return False

        message_sent = False
        for i in range(len(self.possible_agents)):
            if actions[i] == NOSEND:
                continue
            if state[i] == FULL:
                message_sent = True
            break

        return message_sent



"""Heuristic policies for Level-Based Foraging env.

Reference:
https://github.com/uoe-agents/lb-foraging/blob/master/lbforaging/agents/heuristic_agent.py
https://github.com/uoe-agents/BRDiv/blob/master/envs/lb-foraging/lbforaging/agents/heuristic_agent.py

"""
from __future__ import annotations

import random
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, cast

from posggym.agents.policy import Policy, PolicyID, PolicyState
from posggym.agents.utils import action_distributions
from posggym.envs.grid_world.level_based_foraging import (
    LBFAction,
    LBFObs,
    LevelBasedForagingModel,
)

if TYPE_CHECKING:
    from posggym.envs.grid_world.core import Coord


class LBFHeuristicPolicy(Policy[LBFAction, LBFObs]):
    """Heuristic agent for the Level-Based Foraging env.

    This is the abstract Level-Based Foraging env heuristic policy class.
    Concrete implementations must implement the get_action_from_obs method.
    """

    def __init__(
        self, model: LevelBasedForagingModel, agent_id: str, policy_id: PolicyID
    ):
        super().__init__(model, agent_id, policy_id)
        assert model.observation_mode in ("vector", "tuple")
        self._rng = random.Random()

    def reset(self, *, seed: int | None = None):
        super().reset(seed=seed)
        if seed is not None:
            self._rng.seed(seed)

    def get_initial_state(self) -> PolicyState:
        state = super().get_initial_state()
        state["last_obs"] = None
        state["agent_pos"] = None
        state["target_pos"] = None
        return state

    def get_next_state(
        self,
        action: LBFAction | None,
        obs: LBFObs,
        state: PolicyState,
    ) -> PolicyState:
        model = cast(LevelBasedForagingModel, self.model)
        agent_obs, food_obs = model.parse_obs(obs)
        other_agent_obs = [o for o in agent_obs[1:] if o[0] > -1]
        return {
            "last_obs": obs,
            "agent_pos": agent_obs[0][:2],
            "target_pos": self._get_target_pos(
                agent_obs[0], food_obs, other_agent_obs, action, state["target_pos"]
            ),
        }

    def sample_action(self, state: PolicyState) -> LBFAction:
        return self.get_pi(state).sample()

    def get_pi(self, state: PolicyState) -> action_distributions.ActionDistribution:
        if state["target_pos"] is None:
            possible_actions = list(LBFAction)
        else:
            possible_actions = self._move_towards(
                state["agent_pos"], state["target_pos"], load_if_adjacent=True
            )
        return action_distributions.DiscreteActionDistribution(
            {a: 1 / len(possible_actions) for a in possible_actions}, self._rng
        )

    def get_value(self, state: PolicyState) -> float:
        raise NotImplementedError(
            f"`get_value()` no implemented by {self.__class__.__name__} policy"
        )

    def _get_target_pos(
        self,
        agent_obs: Tuple[int, int, int],
        food_obs: List[Tuple[int, int, int]],
        other_agent_obs: List[Tuple[int, int, int]],
        last_action: LBFAction,
        target_pos: Optional[Coord],
    ) -> Optional[Coord]:
        """Get target position from observations."""
        raise NotImplementedError

    def _get_food_by_distance(
        self,
        agent_pos: Coord,
        food_obs: List[Tuple[int, int, int]],
        closest: bool = True,
        max_food_level: Optional[int] = None,
    ) -> Optional[Coord]:
        food_distances: Dict[int, List[Coord]] = {}
        for y, x, level in food_obs:
            if x == -1 or (max_food_level is not None and level > max_food_level):
                continue
            distance = (agent_pos[0] - y) ** 2 + (agent_pos[1] - x) ** 2
            if distance not in food_distances:
                food_distances[distance] = []
            food_distances[distance].append((y, x))

        if len(food_distances) == 0:
            # No food in sight
            return None

        desired_dist = min(food_distances) if closest else max(food_distances)
        return self._rng.choice(food_distances[desired_dist])

    def _center_of_agents(self, agent_obs: List[Tuple[int, int, int]]) -> Coord:
        y_mean = sum(o[0] for o in agent_obs) / len(agent_obs)
        x_mean = sum(o[1] for o in agent_obs) / len(agent_obs)
        return round(x_mean), round(y_mean)

    def _move_towards(
        self, agent_pos: Coord, target: Coord, load_if_adjacent: bool = True
    ) -> List[LBFAction]:
        if (
            load_if_adjacent
            and abs(target[0] - agent_pos[0]) + abs(target[1] - agent_pos[1]) == 1
        ):
            return [LBFAction.LOAD]

        valid_actions = []
        # Note positioning is relative to observing agents observation grid, not the
        # global grid. So relative directions are different.
        if target[0] < agent_pos[0]:
            valid_actions.append(LBFAction.WEST)
        elif target[0] > agent_pos[0]:
            valid_actions.append(LBFAction.EAST)

        if target[1] > agent_pos[1]:
            valid_actions.append(LBFAction.SOUTH)
        elif target[1] < agent_pos[1]:
            valid_actions.append(LBFAction.NORTH)

        if not valid_actions:
            valid_actions.append(LBFAction.NONE)
        return valid_actions

    def _get_updated_pos(self, prev_pos: Coord, last_action: LBFAction) -> Coord:
        # Updates relative position based on last action
        if last_action == LBFAction.NORTH:
            return prev_pos[0], prev_pos[1] + 1
        elif last_action == LBFAction.SOUTH:
            return prev_pos[0], prev_pos[1] - 1
        elif last_action == LBFAction.EAST:
            return prev_pos[0] - 1, prev_pos[1]
        elif last_action == LBFAction.WEST:
            return prev_pos[0] + 1, prev_pos[1]
        else:
            return prev_pos


class LBFHeuristic1(LBFHeuristicPolicy):
    """H1 always goes to the closest observed food, irrespective of the foods level."""

    def _get_target_pos(
        self,
        agent_obs: Tuple[int, int, int],
        food_obs: List[Tuple[int, int, int]],
        other_agent_obs: List[Tuple[int, int, int]],
        last_action: LBFAction,
        target_pos: Optional[Coord],
    ) -> Optional[Coord]:
        agent_pos = agent_obs[:2]
        return self._get_food_by_distance(
            agent_pos, food_obs, closest=True, max_food_level=None
        )


class LBFHeuristic2(LBFHeuristicPolicy):
    """H2 goes towards the visible food closest to the centre of visible players,
    irrespective of food level.
    """

    def _get_target_pos(
        self,
        agent_obs: Tuple[int, int, int],
        food_obs: List[Tuple[int, int, int]],
        other_agent_obs: List[Tuple[int, int, int]],
        last_action: LBFAction,
        target_pos: Optional[Coord],
    ) -> Optional[Coord]:
        if not other_agent_obs:
            return None

        center_pos = self._center_of_agents(other_agent_obs)
        return self._get_food_by_distance(
            center_pos, food_obs, closest=True, max_food_level=None
        )


class LBFHeuristic3(LBFHeuristicPolicy):
    """H3 goes towards the closest visible food with a compatible level."""

    def _get_target_pos(
        self,
        agent_obs: Tuple[int, int, int],
        food_obs: List[Tuple[int, int, int]],
        other_agent_obs: List[Tuple[int, int, int]],
        last_action: LBFAction,
        target_pos: Optional[Coord],
    ) -> Optional[Coord]:
        agent_pos, agent_level = agent_obs[:2], agent_obs[2]
        return self._get_food_by_distance(
            agent_pos, food_obs, closest=True, max_food_level=agent_level
        )


class LBFHeuristic4(LBFHeuristicPolicy):
    """H4 selects and goes towards the visible food that is furthest from the center of
    visible players and that is compatible with the agents level.
    """

    def _get_target_pos(
        self,
        agent_obs: Tuple[int, int, int],
        food_obs: List[Tuple[int, int, int]],
        other_agent_obs: List[Tuple[int, int, int]],
        last_action: LBFAction,
        target_pos: Optional[Coord],
    ) -> Optional[Coord]:
        if target_pos is not None:
            # At the start of an episode it will select a target food and move towards
            # it. Each time it's current target food is collected it then selects a new
            # target based on the heuristic above.
            new_target_pos = self._get_updated_pos(target_pos, last_action)
            if new_target_pos in (f[:2] for f in food_obs):
                return new_target_pos

        # select new target
        if not other_agent_obs:
            # act randomly until we see other agents
            return None

        center_pos = self._center_of_agents(other_agent_obs)
        return self._get_food_by_distance(
            center_pos, food_obs, closest=False, max_food_level=agent_obs[2]
        )


class LBFHeuristic5(LBFHeuristicPolicy):
    """H5 targets a random visible food whose level is compatible with all visible
    agents.
    """

    def _get_target_pos(
        self,
        agent_obs: Tuple[int, int, int],
        food_obs: List[Tuple[int, int, int]],
        other_agent_obs: List[Tuple[int, int, int]],
        last_action: LBFAction,
        target_pos: Optional[Coord],
    ) -> Optional[Coord]:
        if target_pos is not None:
            # At the start of an episode it will select a target food and move towards
            # it. Each time it's current target food is collected it then selects a new
            # target based on the heuristic above.
            new_target_pos = self._get_updated_pos(target_pos, last_action)
            if new_target_pos in (f[:2] for f in food_obs):
                return new_target_pos

        level_sum = sum([o[2] for o in other_agent_obs]) + agent_obs[2]
        food_coords = [f[:2] for f in food_obs if f[1] > -1 and f[2] <= level_sum]
        if not food_coords:
            return None
        return self._rng.choice(food_coords)



"""Heuristic policies for the PredatorPrey grid world environment."""
from __future__ import annotations

from typing import TYPE_CHECKING, List, Tuple, cast

import posggym.envs.grid_world.predator_prey as pp
from posggym.agents.policy import Policy, PolicyID, PolicyState
from posggym.agents.utils import action_distributions
from posggym.envs.grid_world.core import Direction
from posggym.utils import seeding

if TYPE_CHECKING:
    from posggym.envs.grid_world.core import Coord
    from posggym.model import POSGModel
    from posggym.utils.history import AgentHistory


DIR_TO_ACTION = [pp.UP, pp.RIGHT, pp.DOWN, pp.LEFT]


class PPHeuristicPolicy(Policy[pp.PPAction, pp.PPObs]):
    """Base class for PredatorPrey environment heuristic policies."""

    VALID_EXPLORE_STRATEGIES = [
        "uniform_random",
        "spiral",
    ]

    def __init__(
        self,
        model: POSGModel,
        agent_id: str,
        policy_id: PolicyID,
        explore_strategy: str = "uniform_random",
        explore_epsilon: float = 0.05,
    ):
        super().__init__(model, agent_id, policy_id)
        assert explore_strategy in self.VALID_EXPLORE_STRATEGIES
        assert 0 <= explore_epsilon <= 1
        self.explore_strategy = explore_strategy
        self.explore_epsilon = explore_epsilon
        self.model = cast(pp.PredatorPreyModel, model)
        self._grid = self.model.grid
        self._action_space = list(range(self.model.action_spaces[agent_id].n))
        self._rng, _ = seeding.std_random()

        # obs dim is # cells in each direction from agent
        self.obs_dim = self.model.obs_dim
        self.obs_width = (2 * self.obs_dim) + 1
        self.agent_obs_idx = (self.obs_dim * self.obs_width) + self.obs_dim
        self.agent_obs_coord = (self.obs_dim, self.obs_dim)

    def reset(self, *, seed: int | None = None):
        super().reset(seed=seed)
        if seed is not None:
            self._rng, _ = seeding.std_random(seed=seed)

    def get_initial_state(self) -> PolicyState:
        state = super().get_initial_state()
        state.update({"pi": None, "last_obs": None, "last_explore_dir": None})
        return state

    def get_next_state(
        self,
        action: pp.PPAction | None,
        obs: pp.PPObs,
        state: PolicyState,
    ) -> PolicyState:
        pred_coords, prey_coords, wall_obs = self.parse_obs(obs)
        actions = self.get_actions_from_obs(pred_coords, prey_coords)

        explore_dir = state["last_explore_dir"]
        if len(actions) == 0:
            actions, explore_dir = self.get_explore_actions_from_obs(
                wall_obs, explore_dir
            )

        action_probs = {}
        for a in self._action_space:
            if a in actions:
                prob = (1 - self.explore_epsilon) / len(actions)
            else:
                prob = self.explore_epsilon / (len(self._action_space) - len(actions))
            action_probs[a] = prob
        pi = action_distributions.DiscreteActionDistribution(action_probs, self._rng)

        return {
            "pi": pi,
            "last_obs": obs,
            "last_explore_dir": explore_dir,
        }

    def get_state_from_history(self, history: AgentHistory) -> PolicyState:
        _, last_obs = history.get_last_step()
        if last_obs is not None:
            return self.get_next_state(None, last_obs, self.get_initial_state())
        return self.get_initial_state()

    def sample_action(self, state: PolicyState) -> pp.PPAction:
        return self.get_pi(state).sample()

    def get_pi(self, state: PolicyState) -> action_distributions.ActionDistribution:
        if state["pi"] is None:
            raise ValueError(
                "Policy state does not contain a valid action distribution. Make sure"
                "to call `step` or `get_next_state` before calling `get_pi`"
            )
        return state["pi"]

    def get_value(self, state: PolicyState) -> float:
        raise NotImplementedError(
            f"`get_value()` no implemented by {self.__class__.__name__} policy"
        )

    def get_actions_from_obs(
        self, pred_coords: List[Coord], prey_coords: List[Coord]
    ) -> List[pp.PPAction]:
        raise NotImplementedError(
            f"`get_pi_from_obs()` not implemented by {self.__class__.__name__} policy"
        )

    def get_explore_actions_from_obs(
        self, wall_obs: List[bool], explore_dir: Direction | None
    ) -> Tuple[List[pp.PPAction], Direction | None]:
        # using list slice for quick copy of list of primitives
        if self.explore_strategy == "uniform_random":
            # random explore
            return pp.ACTIONS[:], None

        if self.explore_strategy == "spiral":
            # N->E E->S S->W  W->N (i.e clockwise, +1 mod 4)
            if explore_dir is None:
                explore_dir = (
                    Direction.EAST if wall_obs[Direction.SOUTH] else Direction.WEST
                )
            for _ in range(4):
                if not wall_obs[explore_dir]:
                    return [DIR_TO_ACTION[explore_dir]], explore_dir
                explore_dir = (1 + explore_dir) % 4

            # walls observed in all directions, just move in any direction
            return pp.ACTIONS[:], None

        raise NotImplementedError(
            f"Explore strategy `{self.explore_strategy}` not implemented by "
            f"{self.__class__.__name__} policy"
        )

    def parse_obs(self, obs: pp.PPObs) -> Tuple[List[Coord], List[Coord], List[bool]]:
        """Parse obs into list of predator coords, prey coords, and wall directions."""
        pred_coords = []
        prey_coords = []
        walls_obs = [False] * 4
        for idx, cell_obs in enumerate(obs):
            if idx == self.agent_obs_idx:
                continue
            col, row = idx % self.obs_width, idx // self.obs_width
            if cell_obs == pp.PREDATOR:
                pred_coords.append((col, row))
            elif cell_obs == pp.PREY:
                prey_coords.append((col, row))
            elif cell_obs == pp.WALL and row == self.agent_obs_coord[1]:
                if col < self.agent_obs_coord[0]:
                    walls_obs[Direction.WEST] = True
                elif col > self.agent_obs_coord[0]:
                    walls_obs[Direction.EAST] = True
            elif cell_obs == pp.WALL and col == self.agent_obs_coord[0]:
                if row < self.agent_obs_coord[1]:
                    walls_obs[Direction.NORTH] = True
                elif row > self.agent_obs_coord[1]:
                    walls_obs[Direction.SOUTH] = True
        return pred_coords, prey_coords, walls_obs

    def get_closest_coord(self, origin: Coord, coords: List[Coord]) -> Coord | None:
        """Get coord of from list that is closest to the origin coord."""
        return min(
            coords,
            key=lambda coord: self._grid.manhattan_dist(origin, coord),
        )

    def get_actions_towards_target(self, target_coord: Coord) -> List[pp.PPAction]:
        """Get action towards target coord."""
        agent_coord = self.agent_obs_coord

        # Note positioning is relative to observing agents observation grid, not the
        # global grid. So relative directions are different.
        actions = []
        if agent_coord[0] < target_coord[0]:
            actions.append(pp.RIGHT)
        elif agent_coord[0] > target_coord[0]:
            actions.append(pp.LEFT)

        if agent_coord[1] < target_coord[1]:
            actions.append(pp.DOWN)
        elif agent_coord[1] > target_coord[1]:
            actions.append(pp.UP)

        return actions


class PPHeuristic1(PPHeuristicPolicy):
    """H1 moves towards closest observed prey, closest observed predator, or explores
    randomly, in that order.
    """

    def __init__(self, model: POSGModel, agent_id: str, policy_id: PolicyID):
        super().__init__(model, agent_id, policy_id, "uniform_random")

    def get_actions_from_obs(
        self, pred_coords: List[Coord], prey_coords: List[Coord]
    ) -> List[pp.PPAction]:
        if len(prey_coords) != 0:
            closest_prey_coord = self.get_closest_coord(
                self.agent_obs_coord, prey_coords
            )
            return self.get_actions_towards_target(closest_prey_coord)

        if len(pred_coords) != 0:
            closest_pred_coord = self.get_closest_coord(
                self.agent_obs_coord, pred_coords
            )
            return self.get_actions_towards_target(closest_pred_coord)

        return []


class PPHeuristic2(PPHeuristicPolicy):
    """H2 moves towards closest observed prey, closest observed predator, or explores in
    a clockwise spiral around arena, in that order.
    """

    def __init__(self, model: POSGModel, agent_id: str, policy_id: PolicyID):
        super().__init__(model, agent_id, policy_id, "spiral")

    def get_actions_from_obs(
        self, pred_coords: List[Coord], prey_coords: List[Coord]
    ) -> List[pp.PPAction]:
        if len(prey_coords) != 0:
            closest_prey_coord = self.get_closest_coord(
                self.agent_obs_coord, prey_coords
            )
            return self.get_actions_towards_target(closest_prey_coord)

        if len(pred_coords) != 0:
            closest_pred_coord = self.get_closest_coord(
                self.agent_obs_coord, pred_coords
            )
            return self.get_actions_towards_target(closest_pred_coord)

        return []


class PPHeuristic3(PPHeuristicPolicy):
    """H3 moves towards closest observed prey to the closest observed predator or
    explores in a clockwise spiral around arena, in that order.
    """

    def __init__(self, model: POSGModel, agent_id: str, policy_id: PolicyID):
        super().__init__(model, agent_id, policy_id, "spiral")

    def get_actions_from_obs(
        self, pred_coords: List[Coord], prey_coords: List[Coord]
    ) -> List[pp.PPAction]:
        if len(prey_coords) == 0:
            return []

        if len(pred_coords) == 0:
            # no predators observed, just move towards closest prey
            closest_prey_coord = self.get_closest_coord(
                self.agent_obs_coord, prey_coords
            )
            return self.get_actions_towards_target(closest_prey_coord)

        # Move towards prey closest to predator
        closest_pred_coord = self.get_closest_coord(self.agent_obs_coord, pred_coords)
        closest_prey_coord = self.get_closest_coord(closest_pred_coord, prey_coords)
        return self.get_actions_towards_target(closest_prey_coord)



"""PyTorch Policies."""
from __future__ import annotations

import os
import pickle
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    NamedTuple,
    Optional,
    Tuple,
    Type,
    Union,
)

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from gymnasium import spaces
from torch.distributions import Categorical, Normal

from posggym import logger
from posggym.agents.policy import ActType, ObsType, Policy, PolicyID, PolicyState
from posggym.agents.registration import PolicySpec
from posggym.agents.utils import action_distributions, processors
from posggym.agents.utils.download import download_from_repo
from posggym.utils import seeding


if TYPE_CHECKING:
    import posggym.model as M


class PPOTorchModelSaveFileFormat(NamedTuple):
    """Format for saving and loading POSGGym PPOLSTMModel."""

    weights: Dict[str, Any]
    trunk_sizes: List[int]
    lstm_size: int
    lstm_layers: int
    head_sizes: List[int]
    activation: str
    lstm_use_prev_action: bool
    lstm_use_prev_reward: bool


class PPOLSTMModel(nn.Module):
    """A PPO LSTM Model.

    Follows architecture used by Rllib, namely:

      Fully Connected Layers -> LSTM Layer -> policy head (Fully connected layer)
                                           -> value head (Fully connected layer)

    Note, this model is designed for the intended use of having it's weights loaded
    from a file, and not for training. Thus it makes no effort to do things like using
    good weight initialization, rather it just aims to match the architecture of the
    saved model's weights.

    Additionally, it's forward function is designed to take in a single step of
    observations (along with lstm state and optional action and reward) and output
    next actions. This could be a single observation or a batch. But does
    not include logic for unrolling a sequence of observations.

    """

    def __init__(
        self,
        obs_space: spaces.Space,
        action_space: spaces.Space,
        trunk_sizes: List[int],
        lstm_size: int,
        lstm_layers: int,
        head_sizes: List[int],
        activation: str,
        lstm_use_prev_action: bool,
        lstm_use_prev_reward: bool,
    ):
        assert isinstance(obs_space, spaces.Box) and len(obs_space.shape) == 1, (
            "Only 1D Box observation spaces are supported for PPO PyTorch Policy "
            "models. Look into using `gymansium.spaces.flatten_space` to flatten your "
            "observation space, along with using the `FlattenPreprocessor` from "
            f"`posggym.agents.utils.preprocessors`. Got {obs_space}."
        )
        super().__init__()
        self.obs_space = obs_space
        self.action_space = action_space

        self.use_prev_action = lstm_use_prev_action
        self.use_prev_reward = lstm_use_prev_reward

        if isinstance(action_space, spaces.Discrete):
            self.action_dim = action_space.n
        elif isinstance(action_space, spaces.MultiDiscrete):
            # Output is a one hot encoding for each action dimension
            # All one hot encodings concatenated together
            # Hence why we sum the action space nvec
            self.action_dim = np.sum(action_space.nvec)
        elif isinstance(action_space, spaces.Box) and action_space.shape is not None:
            # Output is mean and log_std for each action dimension
            # Hence why we multiply by 2
            # First half of output is mean, second half is log_std
            self.action_dim = np.prod(action_space.shape, dtype=np.int32) * 2
        else:
            raise ValueError(
                f"Unsupported action space for PPOLSTMModel `{type(action_space)}`. "
                "Expected either Discrete, MultiDiscrete, or Box."
            )

        activation_fn: Optional[Callable] = None
        if activation == "tanh":
            activation_fn = nn.Tanh
        elif activation == "relu":
            activation_fn = nn.ReLU

        # Fully connected trunk
        prev_size = int(np.product(obs_space.shape))
        trunk = []
        for size in trunk_sizes:
            trunk.append(nn.Linear(prev_size, size))
            if activation_fn:
                trunk.append(activation_fn())
            prev_size = size
        self.trunk = nn.Sequential(*trunk)

        # LSTM Layer
        lstm_input_size = prev_size
        if self.use_prev_action:
            lstm_input_size += self.action_dim
        if self.use_prev_reward:
            lstm_input_size += 1

        self.lstm = nn.LSTM(
            lstm_input_size, lstm_size, num_layers=lstm_layers, batch_first=False
        )
        prev_size = lstm_size

        # Fully connected actor and critic heads
        actor = []
        critic = []
        for size in head_sizes:
            actor.append(nn.Linear(prev_size, size))
            critic.append(nn.Linear(prev_size, size))
            if activation_fn:
                actor.append(activation_fn())
                critic.append(activation_fn())
            prev_size = size

        actor.append(nn.Linear(prev_size, self.action_dim))
        critic.append(nn.Linear(prev_size, 1))
        self.actor = nn.Sequential(*actor)
        self.critic = nn.Sequential(*critic)

        # Assume model is used for evaluation only
        self.eval()

    @property
    def device(self) -> torch.device:
        """The device this model is on."""
        return next(self.parameters()).device

    def get_next_state(
        self,
        obs: Union[np.ndarray, torch.Tensor],
        lstm_state: Tuple[torch.Tensor, torch.Tensor],
        prev_action: Optional[Union[np.ndarray, torch.Tensor]] = None,
        prev_reward: Optional[Union[np.ndarray, torch.Tensor]] = None,
    ) -> Tuple[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
        """Get next lstm output and state.

        If obs is not batched, adds batch dimension with batch size of 1.

        Arguments
        ---------
        obs: the observation, shape=(batch_size, obs_size) | (obs_size, )
        lstm_state: the lstm state, this is a tuple of two tensors, each with
            shape=(num_layers, batch_size, cell_size)
        prev_action: the previous actions, shape=(batch_size, action_size) |
            (action_size, )
        prev_reward: the previous reward, shape=(batch_size, 1) | (1, )

        Returns
        -------
        lstm_output: the lstm output, shape=(batch_size, cell_size)
        next_lstm_state: the next lstm state, this is a tuple of two tensors, each with
            shape=(num_layers, batch_size, cell_size)
        """
        if isinstance(obs, np.ndarray):
            obs = torch.tensor(obs, dtype=torch.float32)

        if len(obs.shape) == 1:
            # Single observation
            # Add batch and sequence length dimensions
            obs = obs.reshape(1, 1, -1)
        elif len(obs.shape) == 2:
            # Batch of observations
            # Add sequence length dimension
            obs = obs.unsqueeze(1)

        hidden = self.trunk(obs)

        prev_action_reward = []
        if self.use_prev_action:
            if isinstance(self.action_space, (spaces.Discrete, spaces.MultiDiscrete)):
                # One-hot encode discrete actions
                prev_action = F.one_hot(
                    torch.tensor(prev_action, dtype=torch.int64),
                    self.action_dim,
                ).float()
            else:
                prev_action = torch.tensor(prev_action, dtype=torch.float32)
            prev_action_reward.append(prev_action)
        if self.use_prev_reward:
            raise ValueError(
                "Using previous reward as input to LSTM layer not currently supported."
            )

        if len(prev_action_reward) > 0:
            hidden = torch.cat([hidden] + prev_action_reward, dim=-1)

        lstm_output, lstm_state = self.lstm(hidden, lstm_state)
        # remove sequence length dimension
        lstm_output = lstm_output.squeeze(0)
        return lstm_output, lstm_state

    def get_value(
        self,
        obs: Union[np.ndarray, torch.Tensor],
        lstm_state: Tuple[torch.Tensor, torch.Tensor],
        prev_action: Optional[Union[np.ndarray, torch.Tensor]] = None,
        prev_reward: Optional[Union[np.ndarray, torch.Tensor]] = None,
    ) -> torch.Tensor:
        """Get value function output.

        If input is not batched, then adds batch dimension with batch size of 1.

        Arguments
        ---------
        obs: the observation, shape=(batch_size, obs_size) | (obs_size, )
        lstm_state: the lstm state, this is a tuple of two tensors, each with
            shape=(num_layers, batch_size, cell_size)
        prev_action: the previous actions, shape=(batch_size, action_size) |
            (action_size, )
        prev_reward: the previous reward, shape=(batch_size, 1) | (1, )

        Returns
        -------
        value: output of value function, shape=(batch_size, 1)

        """
        hidden_state, _ = self.get_next_state(obs, lstm_state, prev_action, prev_reward)
        return self.critic(hidden_state)

    def get_action_and_value(
        self,
        obs: Union[np.ndarray, torch.Tensor],
        lstm_state: Tuple[torch.Tensor, torch.Tensor],
        prev_action: Optional[Union[np.ndarray, torch.Tensor]] = None,
        prev_reward: Optional[Union[np.ndarray, torch.Tensor]] = None,
        deterministic: bool = False,
    ) -> Tuple[
        torch.Tensor, Tuple[torch.Tensor, torch.Tensor], torch.Tensor, torch.Tensor
    ]:
        """Get next action and value.

        If input is not batched, then adds batch dimension with batch size of 1.

        Arguments
        ---------
        obs: the observation, shape=(batch_size, obs_size) | (obs_size, )
        lstm_state: the lstm state, this is a tuple of two tensors, each
            with shape=(num_layers, batch_size, cell_size)
        prev_action: the previous actions, shape=(batch_size, action_size) |
            (action_size, )
        prev_reward: the previous reward, shape=(batch_size, 1) | (1, )
        deterministic: whether to sample action from action distribution or
            deterministicly select action with highest probability.

        Returns
        -------
        action: next action, shape=(batch_size, action_size)
        next_lstm_state: state of LSTM layer after processing input, this is a tuple of
            two tensors, each with shape=(num_layers, batch_size, cell_size)
        value: output of value function, shape=(batch_size, 1)
        action_dist: the policy action distribution, shape will differ depending on
            the action space.

        """
        with torch.no_grad():
            hidden_state, next_lstm_state = self.get_next_state(
                obs, lstm_state, prev_action, prev_reward
            )
            value = self.critic(hidden_state)
            logits = self.actor(hidden_state)
            if isinstance(self.action_space, spaces.Discrete):
                action_dist = Categorical(logits=logits)
                if deterministic:
                    action = action_dist.probs.argmax(dim=1)
                else:
                    action = action_dist.sample()
                probs = action_dist.probs
            elif isinstance(self.action_space, spaces.MultiDiscrete):
                logit_splits = logits.split(self.action_space.nvec.tolist(), dim=1)
                action_dists = [Categorical(logits=split) for split in logit_splits]
                if deterministic:
                    actions = [dist.probs.argmax(dim=1) for dist in action_dists]
                else:
                    actions = [dist.sample() for dist in action_dists]
                action = torch.stack(actions, dim=1)
                probs = torch.stack([dist.probs for dist in action_dists], dim=1)
            else:
                # Box - continuous action space
                mean, log_std = logits.chunk(2, dim=1)
                action_dist = Normal(mean, log_std.exp())
                action = action_dist.mean if deterministic else action_dist.sample()
                probs = torch.stack([action_dist.mean, action_dist.stddev])

        return action, next_lstm_state, value, probs

    def get_initial_state(
        self, batch_size: int = 1
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Get the initial LSTM state.

        Arguments
        ---------
        batch_size: the batch size of the LSTM state

        Returns
        -------
        initial_state: the initial LSTM state, this is a tuple of two tensors, each
            with shape=(num_layers, batch_size, cell_size)

        """
        return (
            torch.zeros((self.lstm.num_layers, batch_size, self.lstm.hidden_size)),
            torch.zeros((self.lstm.num_layers, batch_size, self.lstm.hidden_size)),
        )


class PPOPolicy(Policy[ActType, ObsType]):
    """A PyTorch PPO Policy.

    Arguments
    ---------
    model: the model of the environment
    agent_id: ID of the agent in the environment the policy is for
    policy_id: ID of the policy
    policy_model: the underlying PyTorch policy model
    obs_processor: the observation processor to use for processing observations before
        they are passed into the policy model. If None, then an identity processor is
        used.
    action_processor: the action processor to use for processing actions before they
        are passed into the policy model, and for unprocessing actions sampled from the
        policy model. If None, then an identity processor is used.
    deterministic: whether to sample actions from the policy model stochastically or
        deterministically. If True, then actions are sampled deterministically.

    """

    def __init__(
        self,
        model: M.POSGModel,
        agent_id: str,
        policy_id: PolicyID,
        policy_model: PPOLSTMModel,
        obs_processor: processors.Processor | None = None,
        action_processor: processors.Processor | None = None,
        deterministic: bool = False,
    ):
        self.policy_model = policy_model
        self.deterministic = deterministic
        self.action_space = model.action_spaces[agent_id]

        if obs_processor is None:
            obs_processor = processors.IdentityProcessor(
                model.observation_spaces[agent_id]
            )
        self.obs_processor = obs_processor

        if action_processor is None:
            action_processor = processors.IdentityProcessor(
                model.action_spaces[agent_id]
            )
        self.action_processor = action_processor

        super().__init__(model, agent_id, policy_id)

        # RNG for sampling actions
        self._rng, _ = seeding.np_random()

    def reset(self, *, seed: int | None = None):
        super().reset(seed=seed)
        if seed is not None:
            # RNG for sampling actions
            self._rng, _ = seeding.np_random(seed=seed)

            import torch

            # https://pytorch.org/docs/stable/notes/randomness.html
            torch.manual_seed(seed)
            torch.use_deterministic_algorithms(True)

            if (
                hasattr(self.policy_model, "device")
                and str(self.policy_model.device) != "cpu"
            ):
                torch.backends.cudnn.benchmark = False
                # See https://github.com/pytorch/pytorch/issues/47672.
                # https://docs.nvidia.com/cuda/cublas/index.html#cublasApi_reproducibility
                cuda_version = torch.version.cuda
                if (
                    cuda_version is not None
                    and float(cuda_version) >= 10.2
                    and (
                        "CUBLAS_WORKSPACE_CONFIG" not in os.environ
                        or (
                            os.environ["CUBLAS_WORKSPACE_CONFIG"]
                            not in (":16:8", ":4096:2")
                        )
                    )
                ):
                    os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":4096:2"

    def get_initial_state(self) -> PolicyState:
        state = super().get_initial_state()
        state["obs"] = None
        state["prev_action"] = None
        state["prev_reward"] = None
        state["lstm_state"] = self.policy_model.get_initial_state(batch_size=1)
        state["action_probs"] = None
        state["value"] = 0.0
        return state

    def get_next_state(
        self, action: ActType | None, obs: ObsType, state: PolicyState
    ) -> PolicyState:
        obs = self.obs_processor(obs)

        if action is not None:
            action = self.action_processor(action)

        (
            _,
            lstm_state,
            value,
            action_probs,
        ) = self.policy_model.get_action_and_value(
            obs,
            state["lstm_state"],
            prev_action=action,
            prev_reward=None,
            deterministic=self.deterministic,
        )

        return {
            "obs": obs,
            "prev_action": action,
            "prev_reward": None,
            "lstm_state": lstm_state,
            "action_probs": action_probs.numpy().squeeze(),
            "value": value[0],
        }

    def sample_action(self, state: PolicyState) -> ActType:
        pi = self.get_pi(state)
        action = pi.sample()
        return self.action_processor.unprocess(action)

    def get_value(self, state: PolicyState) -> float:
        return state["value"]

    def get_pi(self, state: PolicyState) -> action_distributions.ActionDistribution:
        processed_action_space = self.action_processor.input_space
        if isinstance(processed_action_space, spaces.Discrete):
            probs = {
                a: state["action_probs"][a] for a in range(processed_action_space.n)
            }
            return action_distributions.DiscreteActionDistribution(probs, self._rng)
        if isinstance(processed_action_space, spaces.MultiDiscrete):
            probs = [
                {
                    a: state["action_probs"][i][a]
                    for a in range(processed_action_space.nvec[i])
                }
                for i in range(len(processed_action_space))
            ]
            return action_distributions.MultiDiscreteActionDistribution(
                probs, self._rng
            )
        return action_distributions.NormalActionDistribution(
            state["action_probs"][0], state["action_probs"][1], self._rng
        )

    @staticmethod
    def load_from_path(
        model: M.POSGModel,
        agent_id: str,
        policy_id: str,
        policy_file_path: str,
        deterministic: bool = False,
        obs_processor_cls: Type[processors.Processor] | None = None,
        obs_processor_config: Dict[str, Any] | None = None,
        action_processor_cls: Type[processors.Processor] | None = None,
        action_processor_config: Dict[str, Any] | None = None,
    ) -> PPOPolicy:
        if not os.path.exists(policy_file_path):
            logger.info(
                f"Local copy of policy file for policy `{policy_id}` not found, so "
                "downloading it from posggym-agents repo and storing local copy for "
                "future use."
            )
            download_from_repo(policy_file_path)

        obs_space = model.observation_spaces[agent_id]
        if obs_processor_cls is None:
            if isinstance(obs_space, spaces.Box) and len(obs_space.shape) == 1:
                obs_processor_cls = processors.IdentityProcessor
                obs_processor_config = None
            else:
                obs_processor_cls = processors.FlattenProcessor
                obs_processor_config = None

        if obs_processor_config is not None:
            obs_processor: processors.Processor = obs_processor_cls(
                obs_space, **obs_processor_config
            )
        else:
            obs_processor = obs_processor_cls(obs_space)

        action_space = model.action_spaces[agent_id]
        if action_processor_cls is None:
            action_processor_cls = processors.IdentityProcessor
            action_processor_config = None

        if action_processor_config is not None:
            action_processor: processors.Processor = action_processor_cls(
                action_space, **action_processor_config
            )
        else:
            action_processor = action_processor_cls(action_space)

        with open(policy_file_path, "rb") as f:
            data = PPOTorchModelSaveFileFormat(**pickle.load(f))

        policy_model = PPOLSTMModel(
            obs_space=obs_processor.get_processed_space(),
            action_space=action_processor.get_processed_space(),
            trunk_sizes=data.trunk_sizes,
            lstm_size=data.lstm_size,
            lstm_layers=data.lstm_layers,
            head_sizes=data.head_sizes,
            activation=data.activation,
            lstm_use_prev_action=data.lstm_use_prev_action,
            lstm_use_prev_reward=data.lstm_use_prev_reward,
        )
        policy_model.load_state_dict(data.weights)

        return PPOPolicy(
            model=model,
            agent_id=agent_id,
            policy_id=policy_id,
            policy_model=policy_model,
            obs_processor=obs_processor,
            action_processor=action_processor,
            deterministic=deterministic,
        )

    @staticmethod
    def get_spec_from_path(
        policy_file_path: str,
        env_id: str,
        env_args: Dict[str, Any] | None,
        env_args_id: str | None = None,
        version: int = 0,
        valid_agent_ids: List[str] | None = None,
        nondeterministic: bool = False,
        description: str | None = None,
        **kwargs,
    ) -> PolicySpec:
        """Load PPO policy spec from policy file.

        Arguments
        ---------
        policy_file_path: path to the policy file.
        env_id: ID of the posggym environment that the policy is for.
        env_args: Optional keywords arguments for the environment that the policy is
            for (if it is a environment specific policy). If None then assumes policy
            can be used for the environment with any arguments.
        env_args_id: Optional ID for the environment arguments. If None then an ID will
            be generated automatically from the env_args.
        version: the policy version
        valid_agent_ids: Optional AgentIDs for agents in environment that policy is
            compatible with. If None then assumes policy can be used for any agent in
            the environment.
        nondeterministic: Whether this policy is non-deterministic even after seeding.
        kwargs: Additional kwargs, if any, to pass to the agent initializing function
        description: Optional description of the policy.

        Returns
        -------
        spec: Policy specs for PPO Policy loaded from policy file.

        """
        # remove file extension
        policy_name = os.path.basename(policy_file_path).split(".")[0]
        kwargs = kwargs.copy()
        kwargs["policy_file_path"] = policy_file_path
        return PolicySpec(
            policy_name=policy_name,
            entry_point=PPOPolicy.load_from_path,  # type: ignore
            version=version,
            env_id=env_id,
            env_args=env_args,
            env_args_id=env_args_id,
            valid_agent_ids=valid_agent_ids,
            nondeterministic=nondeterministic,
            kwargs=kwargs,
            description=description,
        )



"""Policies for the PredatorPreyContinuous-v0 environment."""
import os.path as osp

from posggym.agents.registration import PolicySpec
from posggym.agents.continuous.predator_prey_continuous import heuristic
from posggym.agents.torch_policy import PPOPolicy
from posggym.config import AGENT_MODEL_DIR
from posggym.agents.utils import processors


ENV_ID = "PredatorPreyContinuous-v0"
agent_model_dir = osp.join(AGENT_MODEL_DIR, "continuous", "predator_prey_continuous")
policy_specs = {}


for n, policy_class in enumerate(
    [
        heuristic.PPCHeuristic0Policy,
        heuristic.PPCHeuristic1Policy,
        heuristic.PPCHeuristic2Policy,
        heuristic.PPCHeuristic3Policy,
    ]
):
    _heuristic_spec = PolicySpec(
        policy_name=f"heuristic{n}",
        entry_point=policy_class,
        version=0,
        env_id=ENV_ID,
        env_args=None,
        valid_agent_ids=None,
        nondeterministic=False,
    )
    policy_specs[_heuristic_spec.id] = _heuristic_spec

# 10x10, 2 predators, 3 prey, prey_strength=2, cooperative=True
for policy_file_name in [
    "sp_seed0.pkl",
    "sp_seed1.pkl",
    "sp_seed2.pkl",
    "sp_seed3.pkl",
    "sp_seed4.pkl",
]:
    spec = PPOPolicy.get_spec_from_path(
        env_id=ENV_ID,
        env_args={
            "world": "10x10",
            "num_predators": 2,
            "num_prey": 3,
            "prey_strength": 2,
            "cooperative": True,
            "obs_dist": 4.0,
            "n_sensors": 16,
        },
        policy_file_path=osp.join(
            agent_model_dir,
            "world=10x10-num_predators=2-num_prey=3-prey_strength=2_cooperative=True-obs_dist=4-n_sensors=16",
            policy_file_name,
        ),
        version=0,
        valid_agent_ids=None,
        # policy is deterministic given random seed
        nondeterministic=False,
        # actions sampled, rather than always taking most probable action
        deterministic=False,
        obs_processor_cls=processors.RescaleProcessor,
        obs_processor_config={"min_val": -1, "max_val": 1},
        action_processor_cls=processors.RescaleProcessor,
        action_processor_config={"min_val": -1, "max_val": 1, "clip": True},
    )
    policy_specs[spec.id] = spec


# 10x10, 4 predators, 3 prey, prey_strength=3, cooperative=True
for policy_file_name in [
    "sp_seed0.pkl",
    "sp_seed1.pkl",
    "sp_seed2.pkl",
    "sp_seed3.pkl",
    "sp_seed4.pkl",
]:
    spec = PPOPolicy.get_spec_from_path(
        env_id=ENV_ID,
        env_args={
            "world": "10x10",
            "num_predators": 4,
            "num_prey": 3,
            "prey_strength": 3,
            "cooperative": True,
            "obs_dist": 4.0,
            "n_sensors": 16,
        },
        policy_file_path=osp.join(
            agent_model_dir,
            "world=10x10-num_predators=4-num_prey=3-prey_strength=3_cooperative=True-obs_dist=4-n_sensors=16",
            policy_file_name,
        ),
        version=0,
        valid_agent_ids=None,
        # policy is deterministic given random seed
        nondeterministic=False,
        # actions sampled, rather than always taking most probable action
        deterministic=False,
        obs_processor_cls=processors.RescaleProcessor,
        obs_processor_config={"min_val": -1, "max_val": 1},
        action_processor_cls=processors.RescaleProcessor,
        action_processor_config={"min_val": -1, "max_val": 1, "clip": True},
    )
    policy_specs[spec.id] = spec



"""Heuristic Bot Policies for the Cooperative Reaching Environment.

Adapted from:
https://github.com/uoe-agents/BRDiv/tree/master/envs/CooperativeReaching/coopreaching/coopreaching/agents

Reference:
Rahman et al (2023) Generating Teammates for Training Robust Ad Hoc Teamwork Agents via
Best-Response Diversity https://openreview.net/pdf?id=l5BzfQhROl

"""
from __future__ import annotations

import random
from typing import TYPE_CHECKING, List, Optional

from posggym.agents.policy import Policy, PolicyID, PolicyState
from posggym.agents.utils import action_distributions
from posggym.envs.grid_world.cooperative_reaching import (
    DO_NOTHING,
    DOWN,
    LEFT,
    RIGHT,
    UP,
    CooperativeReachingModel,
    CRAction,
    CRObs,
)

if TYPE_CHECKING:
    from posggym.envs.grid_world.core import Coord


class CRHeuristicPolicy(Policy[CRAction, CRObs]):
    """Heuristic policy for the Cooperative Reaching Environment.

    This is the base class for all heuristic policies. Concrete implementations must
    implement the `_get_target_pos` method.
    """

    def __init__(
        self, model: CooperativeReachingModel, agent_id: str, policy_id: PolicyID
    ):
        super().__init__(model, agent_id, policy_id)
        self._rng = random.Random()
        self.grid_size = model.size
        self.goals = model.goals

    def reset(self, *, seed: int | None = None):
        super().reset(seed=seed)
        if seed is not None:
            self._rng.seed(seed)

    def get_initial_state(self) -> PolicyState:
        state = super().get_initial_state()
        state["last_obs"] = None
        state["target_goal"] = None
        return state

    def get_next_state(
        self,
        action: CRAction | None,
        obs: CRObs,
        state: PolicyState,
    ) -> PolicyState:
        return {
            "last_obs": obs,
            "target_goal": self._get_target_pos(obs, state["target_goal"]),
        }

    def sample_action(self, state: PolicyState) -> CRAction:
        return self.get_pi(state).sample()

    def get_pi(self, state: PolicyState) -> action_distributions.ActionDistribution:
        possible_actions = self._move_towards(
            state["target_goal"], tuple(state["last_obs"][0])
        )
        return action_distributions.DiscreteActionDistribution(
            {a: 1 / len(possible_actions) for a in possible_actions}, self._rng
        )

    def get_value(self, state: PolicyState) -> float:
        raise NotImplementedError(
            f"`get_value()` no implemented by {self.__class__.__name__} policy"
        )

    def _move_towards(self, target_pos: Coord, agent_pos: Coord) -> List[CRAction]:
        """Get list of actions that move towards target_pos from agent_pos."""
        valid_actions = []
        if target_pos[1] < agent_pos[1]:
            valid_actions.append(UP)
        elif target_pos[1] > agent_pos[1]:
            valid_actions.append(DOWN)

        if target_pos[0] < agent_pos[0]:
            valid_actions.append(LEFT)
        elif target_pos[0] > agent_pos[0]:
            valid_actions.append(RIGHT)

        if not valid_actions:
            valid_actions.append(DO_NOTHING)
        return valid_actions

    def _get_target_pos(self, obs: CRObs, target_goal: Optional[Coord]) -> Coord:
        """Get target position from observation."""
        raise NotImplementedError

    def _get_target_goal(
        self,
        obs: CRObs,
        closest: bool | None,
        optimal: bool | None,
    ) -> Coord:
        """Get target goal position from observation.

        Conditions selection on whether to go to the closest (closest=True), furthest
        (closest=False), or any distance (closest=None) goal, and whether to go to the
        optimal (highest value, optimal=True), suboptimal (non-max value, optimal=False)
        or any value (optimal=None) goal.
        """
        agent_pos = obs[0]

        if optimal is None:
            goal_list = list(self.goals)
        elif optimal:
            max_reward = max(self.goals.values())
            goal_list = [
                goal for goal, reward in self.goals.items() if reward == max_reward
            ]
        else:
            max_reward = max(self.goals.values())
            goal_list = [
                goal for goal, reward in self.goals.items() if reward != max_reward
            ]
            if not goal_list:
                # all goals are equal
                goal_list = list(self.goals)

        if closest is None:
            return self._rng.choice(goal_list)

        dist_to_goal = [
            abs(agent_pos[0] - goal[0]) + abs(agent_pos[1] - goal[1])
            for goal in goal_list
        ]
        desired_dist_to_goal = min(dist_to_goal) if closest else max(dist_to_goal)
        valid_target_goals = [
            g
            for g, dist in zip(goal_list, dist_to_goal)
            if dist == desired_dist_to_goal
        ]
        return self._rng.choice(valid_target_goals)


class CRHeuristic1(CRHeuristicPolicy):
    """H1 always goes to the closest rewarding goal."""

    def _get_target_pos(self, obs: CRObs, target_goal: Optional[Coord]) -> Coord:
        if target_goal is None:
            target_goal = self._get_target_goal(obs, closest=True, optimal=None)
        return target_goal


class CRHeuristic2(CRHeuristicPolicy):
    """H2 always goes to the furthest rewarding goal."""

    def _get_target_pos(self, obs: CRObs, target_goal: Optional[Coord]) -> Coord:
        if target_goal is None:
            target_goal = self._get_target_goal(obs, closest=False, optimal=None)
        return target_goal


class CRHeuristic3(CRHeuristicPolicy):
    """H3 always goes to the closest optimal goal."""

    def _get_target_pos(self, obs: CRObs, target_goal: Optional[Coord]) -> Coord:
        if target_goal is None:
            target_goal = self._get_target_goal(obs, closest=True, optimal=True)
        return target_goal


class CRHeuristic4(CRHeuristicPolicy):
    """H4 always goes to the furthest optimal goal."""

    def _get_target_pos(self, obs: CRObs, target_goal: Optional[Coord]) -> Coord:
        if target_goal is None:
            target_goal = self._get_target_goal(obs, closest=False, optimal=True)
        return target_goal


class CRHeuristic5(CRHeuristicPolicy):
    """H5 always goes to the closest suboptimal goal."""

    def _get_target_pos(self, obs: CRObs, target_goal: Optional[Coord]) -> Coord:
        if target_goal is None:
            target_goal = self._get_target_goal(obs, closest=True, optimal=False)
        return target_goal


class CRHeuristic6(CRHeuristicPolicy):
    """H6 always goes to the furthest suboptimal goal."""

    def _get_target_pos(self, obs: CRObs, target_goal: Optional[Coord]) -> Coord:
        if target_goal is None:
            target_goal = self._get_target_goal(obs, closest=False, optimal=False)
        return target_goal


class CRHeuristic7(CRHeuristicPolicy):
    """H7 goes to a randomly selected goal."""

    def _get_target_pos(self, obs: CRObs, target_goal: Optional[Coord]) -> Coord:
        if target_goal is None:
            target_goal = self._rng.choice(list(self.goals))
        return target_goal


class CRHeuristic8(CRHeuristicPolicy):
    """H8 goes to the goal closest to the other agent at each time step."""

    def _get_target_pos(self, obs: CRObs, target_goal: Optional[Coord]) -> Coord:
        other_pos = obs[1]
        if other_pos == (self.grid_size, self.grid_size):
            # cannot see other agent, so just go towards a random goal
            return self._rng.choice(list(self.goals))

        dist_to_goal = [
            abs(other_pos[0] - goal[0]) + abs(other_pos[1] - goal[1])
            for goal in self.goals
        ]
        min_dist_to_goal = min(dist_to_goal)
        closest_goals = [
            g for g, dist in zip(self.goals, dist_to_goal) if dist == min_dist_to_goal
        ]
        return self._rng.choice(closest_goals)


class CRHeuristic9(CRHeuristicPolicy):
    """H9 goes to the optimal goal closest to the other agent."""

    def _get_target_pos(self, obs: CRObs, target_goal: Optional[Coord]) -> Coord:
        other_pos = obs[1]
        if other_pos == (self.grid_size, self.grid_size):
            # cannot see other agent, so just go towards a random optimal goal
            return self._get_target_goal(obs, closest=None, optimal=True)

        goal_list = [
            goal
            for goal, reward in self.goals.items()
            if reward == max(self.goals.values())
        ]

        dist_to_goal = [
            abs(other_pos[0] - goal[0]) + abs(other_pos[1] - goal[1])
            for goal in goal_list
        ]
        min_dist_to_goal = min(dist_to_goal)
        closest_goals = [
            g for g, dist in zip(goal_list, dist_to_goal) if dist == min_dist_to_goal
        ]
        return self._rng.choice(closest_goals)


class CRHeuristic10(CRHeuristicPolicy):
    """H10 goes to the sub-optimal goal closest to the other agent."""

    def _get_target_pos(self, obs: CRObs, target_goal: Optional[Coord]) -> Coord:
        other_pos = obs[1]
        if other_pos == (self.grid_size, self.grid_size):
            # cannot see other agent, so just go towards a random optimal goal
            return self._get_target_goal(obs, closest=None, optimal=True)

        goal_list = [
            goal
            for goal, reward in self.goals.items()
            if reward == min(self.goals.values())
        ]

        dist_to_goal = [
            abs(other_pos[0] - goal[0]) + abs(other_pos[1] - goal[1])
            for goal in goal_list
        ]
        min_dist_to_goal = min(dist_to_goal)
        closest_goals = [
            g for g, dist in zip(goal_list, dist_to_goal) if dist == min_dist_to_goal
        ]
        return self._rng.choice(closest_goals)


class CRHeuristic11(CRHeuristicPolicy):
    """H11 follows the other agent."""

    def _get_target_pos(self, obs: CRObs, target_goal: Optional[Coord]) -> Coord:
        other_pos = obs[1]
        if other_pos == (self.grid_size, self.grid_size):
            # cannot see other agent, so just go towards a random goal
            return self._rng.choice(list(self.goals))
        return other_pos



import os
import os.path as osp
from pathlib import Path


PKG_DIR = osp.dirname(osp.abspath(__file__))
REPO_DIR = osp.abspath(osp.join(PKG_DIR, os.pardir))
BASE_RESULTS_DIR = osp.join(str(Path.home()), "posggym_results")
ASSET_DIR = osp.join(PKG_DIR, "assets")
AGENT_MODEL_DIR = osp.join(ASSET_DIR, "agents")

AGENT_MODEL_REPO_URL = "https://github.com/RDLLab/posggym-agent-models/raw/v0.5.0/"


if not osp.exists(BASE_RESULTS_DIR):
    os.makedirs(BASE_RESULTS_DIR)
